"""Controller holding and managing Afero IoT resources of type `portable-air-conditioner`."""

import copy

from aioafero.device import AferoDevice
from aioafero.errors import DeviceNotFound
from aioafero.v1.models import features
from aioafero.v1.models.portable_ac import PortableAC, PortableACPut
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .climate import ClimateController
from .event import CallbackResponse

SPLIT_IDENTIFIER: str = "portable-ac"


def generate_split_name(afero_device: AferoDevice, instance: str) -> str:
    """Generate the name for an instanced element."""
    return f"{afero_device.id}-{SPLIT_IDENTIFIER}-{instance}"


def get_valid_states(afero_dev: AferoDevice) -> list:
    """Find states associated with the element."""
    return [
        state
        for state in afero_dev.states
        if state.functionClass in ["available", "power"]
    ]


def portable_ac_callback(afero_device: AferoDevice) -> CallbackResponse:
    """Convert an AferoDevice into multiple devices."""
    multi_devs: list[AferoDevice] = []
    if afero_device.device_class == ResourceTypes.PORTABLE_AC.value:
        instance = "power"
        cloned = copy.deepcopy(afero_device)
        cloned.id = generate_split_name(afero_device, instance)
        cloned.split_identifier = SPLIT_IDENTIFIER
        cloned.friendly_name = f"{afero_device.friendly_name} - {instance}"
        cloned.states = get_valid_states(afero_device)
        cloned.device_class = ResourceTypes.SWITCH.value
        cloned.children = []
        multi_devs.append(cloned)
    return CallbackResponse(
        split_devices=multi_devs,
        remove_original=False,
    )


class PortableACController(ClimateController[PortableAC]):
    """Controller holding and managing Afero IoT resources of type `portable-air-conditioner`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.PORTABLE_AC]
    ITEM_CLS = PortableAC
    ITEM_MAPPING = {
        "hvac_mode": "mode",
    }
    # Elements that map to Select. func class / func instance to name
    ITEM_SELECTS = {
        ("fan-speed", "ac-fan-speed"): "Fan Speed",
        ("sleep", None): "Sleep Mode",
        ("air-swing", None): "Swing",
    }
    DEVICE_SPLIT_CALLBACKS: dict[str, callable] = {
        ResourceTypes.PORTABLE_AC.value: portable_ac_callback
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> PortableAC:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        climate_data = await self.initialize_climate_elem(afero_device)

        self._items[afero_device.id] = PortableAC(
            _id=afero_device.id,
            available=climate_data["available"],
            current_temperature=climate_data["current_temperature"],
            hvac_mode=climate_data["hvac_mode"],
            target_temperature_heating=climate_data["target_temperature_heating"],
            target_temperature_cooling=climate_data["target_temperature_cooling"],
            numbers={},
            selects=climate_data["selects"],
            binary_sensors={},
            sensors={},
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                children=afero_device.children,
                functions=afero_device.functions,
            ),
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Portable AC with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        return await self.update_climate_elem(afero_device)

    async def set_state(self, device_id: str, **kwargs) -> None:
        """Set supported feature(s) to portable ac resource."""
        update_obj = PortableACPut()
        hvac_mode: str | None = kwargs.get("hvac_mode")
        target_temperature: float | None = kwargs.get("target_temperature")
        selects: dict[tuple[str, str | None], str] | None = kwargs.get("selects", {})
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if hvac_mode:
            if hvac_mode in cur_item.hvac_mode.supported_modes:
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode=hvac_mode,
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown hvac mode %s. Available modes: %s",
                    hvac_mode,
                    ", ".join(sorted(cur_item.hvac_mode.modes)),
                )
        if target_temperature is not None:
            kwargs["target_temperature_cooling"] = kwargs.pop("target_temperature")

        for key, val in selects.items():
            if key not in cur_item.selects:
                continue
            update_obj.selects[key] = features.SelectFeature(
                selected=val,
                selects=cur_item.selects[key].selects,
                name=cur_item.selects[key].name,
            )
        await self.set_climate_state(device_id, update_obj, **kwargs)



import pytest

from aioafero.v1.models import PortableAC, features, DeviceInformation


@pytest.fixture
def populated_entity():
    numbers = {
        ("timer", None): features.NumbersFeature(
            value=0, min=0, max=1440, step=30, name="timer", unit="minutes"
        )
    }
    selects = {
        ("fan-speed", "ac-fan-speed"): features.SelectFeature(
            selected="fan-speed-auto",
            selects={"fan-speed-auto", "fan-speed-2-100", "fan-speed-2-050"},
            name="Fan Speed",
        ),
        ("sleep", None): features.SelectFeature(
            selected="off",
            selects={"on", "off"},
            name="Sleep Timer",
        ),
    }
    return PortableAC(
        _id="entity-1",
        available=True,
        current_temperature=features.CurrentTemperatureFeature(
            temperature=35,
            function_class="temperature",
            function_instance="current-temp",
        ),
        hvac_mode=features.HVACModeFeature(
            mode="auto-cool",
            previous_mode="fan",
            modes={"fan", "auto-cool", "dehumidify", "cool"},
            supported_modes={"fan", "auto-cool", "dehumidify", "cool"},
        ),
        target_temperature_cooling=features.TargetTemperatureFeature(
            value=26, step=0.5, min=10, max=37, instance="cooling-target"
        ),
        numbers=numbers,
        selects=selects,
        device_information=DeviceInformation(
            functions=[
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ]
        )
    )


def test_init(populated_entity):
    assert populated_entity.id == "entity-1"
    assert populated_entity.available is True
    assert populated_entity.instances == {"preset": "preset-1"}
    assert populated_entity.temperature == 35
    assert populated_entity.current_temperature.function_class == "temperature"
    assert populated_entity.current_temperature.function_instance == "current-temp"
    assert populated_entity.hvac_mode.mode == "auto-cool"
    assert populated_entity.target_temperature_cooling.value == 26
    assert populated_entity.target_temperature_cooling.step == 0.5
    assert populated_entity.target_temperature_cooling.min == 10
    assert populated_entity.target_temperature_cooling.max == 37
    # Property checks
    assert populated_entity.target_temperature == 26
    assert populated_entity.target_temperature_max == 37
    assert populated_entity.target_temperature_min == 10
    assert populated_entity.target_temperature_step == 0.5
    assert populated_entity.supports_fan_mode is False
    assert populated_entity.supports_temperature_range is False


def test_get_instance(populated_entity):
    assert populated_entity.get_instance("preset") == "preset-1"


def test_empty():
    entity = PortableAC(
        _id="entity-1",
        available=True,
        device_information=DeviceInformation(functions=[]),
        numbers={},
        selects={},
    )
    assert entity.get_mode_to_check() is None
    assert entity.temperature is None
    assert entity.target_temperature is None
    assert entity.target_temperature_max is None
    assert entity.target_temperature_min is None
    assert entity.supports_fan_mode is False
    assert entity.supports_temperature_range is False
    assert entity.target_temperature_step == 1


@pytest.mark.parametrize(
    ("mode", "previous_mode", "expected"),
    [
        (None, None, None),
        ("cool", None, "cool"),
        ("dehumidify", "cool", "cool"),
        ("auto-cool", None, "cool"),
        ("heat", "heat", "heat"),
        ("heat", None, "heat"),
        ("fan", None, None),
    ]
)
def test_get_mode_to_check(mode, previous_mode, expected, populated_entity):
    populated_entity.hvac_mode.mode = mode
    populated_entity.hvac_mode.previous_mode = previous_mode
    assert populated_entity.get_mode_to_check() == expected



"""Representation of an Afero Sensor and its corresponding updates."""

from dataclasses import dataclass, field


@dataclass
class AferoSensor:
    """Sensor that can return any value."""

    id: str
    owner: str
    value: str | int | float | None

    unit: str | None = field(default=None)
    instance: str | None = field(default=None)


@dataclass
class AferoBinarySensor:
    """Sensor that can return True (alerting) or False (normal)."""

    id: str
    owner: str
    current_value: str | int
    _error: str | int

    unit: str | None = field(default=None)
    instance: str | None = field(default=None)

    @property
    def value(self) -> bool:
        """Determine if the binary sensor is in an alert state."""
        return self.current_value == self._error



"""Test SecuritySystemController"""

import pytest

from aioafero.v1.controllers.security_system_keypad import SecuritySystemKeypadController, features, AferoBinarySensor
from aioafero import AferoState

from .. import utils


security_system = utils.create_devices_from_data("security-system.json")
keypad_id = "1f31be19-b9b9-4ca8-8a22-20d0015ec2dd"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.security_systems_keypads


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller[keypad_id]
    assert dev.available is True
    assert dev.selects == {
        ("volume", "buzzer-volume"): features.SelectFeature(
            selected="volume-02",
            selects={"volume-00", "volume-01", "volume-02", "volume-03", "volume-04"},
            name="Buzzer Volume",
        ),
    }
    assert dev.binary_sensors == {
        'tamper-detection|None': AferoBinarySensor(
            id='tamper-detection|None',
            owner='ce68c348-cd68-4ce3-937d-bf18619ae970',
            current_value='not-tampered',
            _error='tampered',
            unit=None,
            instance=None)
    }


@pytest.mark.asyncio
async def test_empty_update(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    update = utils.create_devices_from_data("security-system.json")[0]
    updates = await mocked_controller.update_elem(update)
    assert updates == set()


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller[keypad_id]
    assert dev.available
    update = utils.create_devices_from_data("security-system.json")[0]
    new_states = [
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="volume", value="volume-03", lastUpdateTime=0, functionInstance="buzzer-volume"
        ),
        AferoState(
            functionClass="tamper-detection", value="tampered", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(update, state)
    updates = await mocked_controller.update_elem(update)
    await mocked_controller._bridge.async_block_until_done()
    assert not dev.available
    assert updates == {
        "available",
        "select-('volume', 'buzzer-volume')",
        "binary-tamper-detection|None",
    }
    assert dev.selects[("volume", "buzzer-volume")].selected == "volume-03"
    assert dev.binary_sensors["tamper-detection|None"].current_value == "tampered"


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(keypad_id)


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        keypad_id,
        selects={("volume", "buzzer-volume"): "volume-03", ("bad", None): False},
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[keypad_id]
    assert dev.selects[("volume", "buzzer-volume")].selected == "volume-03"

@pytest.mark.asyncio
async def test_set_state_bad_device(mocked_controller):
    await mocked_controller._bridge.generate_devices_from_data(security_system)
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        "doesnt-exist",
        selects={("volume", "buzzer-volume"): "volume-03", ("bad", None): False},
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_emitting(mocked_controller):
    bridge = mocked_controller._bridge
    await bridge.generate_devices_from_data(security_system)
    await bridge.async_block_until_done()
    dev = bridge.security_systems_keypads[keypad_id]
    assert dev.available
    update = utils.create_devices_from_data("security-system.json")[0]
    new_states = [
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="volume", value="volume-03", lastUpdateTime=0, functionInstance="buzzer-volume"
        ),
        AferoState(
            functionClass="tamper-detection", value="tampered", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(update, state)
    await bridge.generate_devices_from_data([update])
    await bridge.async_block_until_done()
    dev = bridge.security_systems_keypads[keypad_id]
    assert not dev.available
    assert dev.selects[("volume", "buzzer-volume")].selected == "volume-03"
    assert dev.binary_sensors["tamper-detection|None"].current_value == "tampered"



"""Generic/base Resource Model(s)."""

from dataclasses import dataclass
from enum import Enum
from typing import Any


class ResourceTypes(Enum):
    """Type of the supported resources."""

    DEVICE = "metadevice.device"
    HOME = "metadata.home"
    ROOM = "metadata.room"
    EXHAUST_FAN = "exhaust-fan"
    FAN = "fan"
    LANDSCAPE_TRANSFORMER = "landscape-transformer"
    LIGHT = "light"
    LOCK = "door-lock"
    PARENT_DEVICE = "parent-device"
    PORTABLE_AC = "portable-air-conditioner"
    POWER_OUTLET = "power-outlet"
    SECURITY_SYSTEM = "security-system"
    SECURITY_SYSTEM_KEYPAD = "security-system-keypad"
    SECURITY_SYSTEM_SENSOR = "security-system-sensor"  # Create from device splits
    SWITCH = "switch"
    THERMOSTAT = "thermostat"
    UNKNOWN = "unknown"
    WATER_TIMER = "water-timer"

    @classmethod
    def _missing_(cls: type, value: object):
        """Set default enum member if an unknown value is provided."""
        return ResourceTypes.UNKNOWN


@dataclass
class DeviceInformation:
    """Generic Device Information."""

    device_class: str | None = None
    default_image: str | None = None
    default_name: str | None = None
    manufacturer: str | None = None
    model: str | None = None
    name: str | None = None
    parent_id: str | None = None
    wifi_mac: str | None = None
    ble_mac: str | None = None
    version_data: dict | None = None
    children: list[str] | None = None
    functions: list[dict[str, Any]] | None = None



import datetime
import json
import logging
import os
import pathlib
import time
from urllib.parse import urlencode

import aiohttp
import pytest

from aioafero.v1 import auth, v1_const

current_path = pathlib.Path(__file__).parent.resolve()


@pytest.fixture(scope="function")
def hs_auth(mocked_bridge_req):
    return auth.AferoAuth(mocked_bridge_req, "username", "password")


async def build_url(base_url: str, qs: dict[str, str]) -> str:
    return f"{base_url}?{urlencode(qs)}"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "time_offset,is_expired",
    [
        # No token
        (None, True),
        # Expired token
        (-5, True),
        # Non-Expired token
        (5, False),
    ],
)
async def test_is_expired(time_offset, is_expired, hs_auth):
    if time_offset:
        hs_auth._token_data = auth.TokenData(
            "token", None, None, time.time() + time_offset
        )
    assert await hs_auth.is_expired == is_expired


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "page_filename,form_id,err_msg,expected",
    [
        # Valid
        (
            "auth_webapp_login.html",
            "kc-form-login",
            None,
            auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id"),
        ),
        # page is missing expected id
        (
            "auth_webapp_login_missing.html",
            "kc-form-login",
            "Unable to parse login page",
            None,
        ),
        # form field is missing expected attribute
        (
            "auth_webapp_login_bad_format.html",
            "kc-form-login",
            "Unable to extract login url",
            None,
        ),
        # URL missing expected elements
        (
            "auth_webapp_login_bad_qs.html",
            "kc-form-login",
            "Unable to parse login url",
            None,
        ),
    ],
)
async def test_extract_login_data(page_filename, form_id, err_msg, expected):
    with open(os.path.join(current_path, "data", page_filename)) as f:
        page_data = f.read()
    if expected:
        assert await auth.extract_login_data(page_data, form_id) == expected
    else:
        with pytest.raises(auth.InvalidResponse, match=err_msg):
            await auth.extract_login_data(page_data, form_id)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "page_filename, gc_exp, redirect, response, expected_err",
    [
        # Invalid status code
        (None, None, False, {"status": 403}, auth.InvalidResponse),
        # Valid auth passed to generate_code
        (
            "auth_webapp_login.html",
            auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id"),
            False,
            {"status": 200},
            None,
        ),
        # Random error
        (
            "auth_webapp_login.html",
            auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id"),
            False,
            {"status": 400},
            auth.InvalidResponse,
        ),
        # Active session returned
        (
            "auth_webapp_login.html",
            None,
            True,
            {
                "status": 302,
                "headers": {
                    "location": (
                        "hubspace-app://loginredirect"
                        "?session_state=sess-state"
                        "&iss=https%3A%2F%2Faccounts.hubspaceconnect.com"
                        "%2Fauth%2Frealms%2Fthd&code=code"
                    )
                },
            },
            None,
        ),
    ],
)
async def test_webapp_login(
    page_filename,
    gc_exp,
    redirect,
    response,
    expected_err,
    hs_auth,
    mock_aioresponse,
    aio_sess,
    mocker,
):
    hs_auth._bridge._web_session = aio_sess
    if page_filename:
        with open(os.path.join(current_path, "data", page_filename)) as f:
            response["body"] = f.read()
    challenge = await hs_auth.generate_challenge_data()
    generate_code = mocker.patch.object(hs_auth, "generate_code")
    parse_code = mocker.patch.object(auth.AferoAuth, "parse_code")
    params: dict[str, str] = {
        "response_type": "code",
        "client_id": v1_const.AFERO_CLIENTS["hubspace"]["AUTH_DEFAULT_CLIENT_ID"],
        "redirect_uri": "hubspace-app%3A%2F%2Floginredirect",
        "code_challenge": challenge.challenge,
        "code_challenge_method": "S256",
        "scope": "openid offline_access",
    }
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_OPENID_ENDPOINT"])
    url = await build_url(url, params)
    mock_aioresponse.get(url, **response)
    if not expected_err:
        await hs_auth.webapp_login(challenge)
        if redirect:
            generate_code.asset_not_called()
            parse_code.assert_called_once()
        else:
            generate_code.assert_called_once_with(gc_exp, challenge)
            parse_code.assert_not_called()
    else:
        with pytest.raises(expected_err):
            await hs_auth.webapp_login(challenge)
        generate_code.assert_not_called()


@pytest.mark.asyncio
async def test_generate_challenge_data():
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "auth_data, response, expected_err, expected",
    [
        # Invalid response
        (
            auth.AuthSessionData("sess_code", "execution", "tab_id"),
            {"status": 200},
            auth.InvalidAuth,
            None,
        ),
        # Invalid Location
        (
            auth.AuthSessionData("sess_code", "execution", "tab_id"),
            {"status": 302, "headers": {"location": "nope"}},
            auth.InvalidResponse,
            None,
        ),
        # Valid location
        (
            auth.AuthSessionData("sess_code", "execution", "tab_id"),
            {"status": 302, "headers": {"location": "https://cool.beans?code=beans"}},
            None,
            "beans",
        ),
        # OTP login required
        (
            auth.AuthSessionData("sess_code", "execution", "tab_id"),
            {"status": 200, "headers": {"location": "https://cool.beans?code=beans"}, "body": '<form id="kc-otp-login-form" class="form-horizontal" action="https://accounts.hubspaceconnect.com/auth/realms/thd/login-actions/authenticate?session_code=session_code&amp;execution=execution&amp;client_id=hubspace_android&amp;tab_id=tab_id" method="post" onsubmit="return submitForm()">'},
            auth.OTPRequired,
            None,
        )
    ],
)
async def test_generate_code(
    auth_data,
    response,
    expected_err,
    expected,
    hs_auth,
    aioresponses,
    aio_sess,
):
    params = {
        "session_code": auth_data.session_code,
        "execution": auth_data.execution,
        "client_id": v1_const.AFERO_CLIENTS["hubspace"]["AUTH_DEFAULT_CLIENT_ID"],
        "tab_id": auth_data.tab_id,
    }
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
    url = await build_url(url, params)
    aioresponses.post(url, **response)
    if not expected_err:
        assert (
            await hs_auth.generate_code(auth_data, None)
            == expected
        )
    else:
        with pytest.raises(expected_err):
            await hs_auth.generate_code(auth_data, None)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "secure_mode,code,response,expected,expected_messages, err",
    [
        # Invalid refresh token
        (True, "code", {"status": 403}, None, None, aiohttp.web_exceptions.HTTPForbidden),
        # Incorrect format
        (
            True,
            "code",
            {"status": 200, "body": json.dumps({"refresh_token2": "cool_beans"})},
            None,
            None,
            auth.InvalidResponse,
        ),
        # Weird stuff returned
        (
            True,
            "code",
            {"status": 400, "body": "{"},
            None,
            None,
            auth.InvalidResponse,
        ),
        # Valid refresh token
        (
            True,
            "those-are-some-cool-beans",
            {
                "status": 200,
                "body": json.dumps(
                    {
                        "id_token": "cool_beans",
                        "refresh_token": "refresh_beans",
                        "access_token": "access_token_beans",
                    }
                ),
            },
            "refresh_beans",
            [
                "data: {'grant_type': 'authorization_code', 'code': 'th***ns'",
                (
                    "JSON response: {'id_token': 'co***ns', 'refresh_token': "
                    "'re***ns', 'access_token': 'ac***ns'}"
                ),
            ],
            None,
        ),
        # Valid refresh token - inseucre
        (
            False,
            "those-are-some-cool-beans",
            {
                "status": 200,
                "body": json.dumps(
                    {
                        "id_token": "cool_beans",
                        "refresh_token": "refresh_beans",
                        "access_token": "access_token_beans",
                    }
                ),
            },
            "refresh_beans",
            [
                (
                    "JSON response: {'id_token': 'cool_beans', 'refresh_token': "
                    "'refresh_beans', 'access_token': 'access_token_beans'}"
                ),
            ],
            None,
        ),
    ],
)
async def test_generate_refresh_token(
    secure_mode,
    code,
    response,
    expected,
    expected_messages,
    err,
    hs_auth,
    aioresponses,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    auth.add_secret("those-are-some-cool-beans")
    if not secure_mode:
        hs_auth.secret_logger = auth.passthrough
    hs_auth._token_data = None
    challenge = await hs_auth.generate_challenge_data()
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_TOKEN_ENDPOINT"])
    aioresponses.post(url, **response)
    if expected:
        assert (
            expected
            == (
                await hs_auth.generate_refresh_token(
                    code=code, challenge=challenge
                )
            ).refresh_token
        )
    else:
        with pytest.raises(err):
            await hs_auth.generate_refresh_token(
                code=code, challenge=challenge
            )
    aioresponses.assert_called_once()
    call_args = list(aioresponses.requests.values())[0][0]
    # Add in the user-agent that is generated from the bridge
    hs_auth._token_headers["user-agent"] = v1_const.AFERO_GENERICS["DEFAULT_USERAGENT"].safe_substitute(
        client_name="aioafero"
    )
    assert call_args.kwargs["headers"] == hs_auth._token_headers
    assert call_args.kwargs["data"] == {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": v1_const.AFERO_CLIENTS["hubspace"]["AUTH_DEFAULT_REDIRECT_URI"],
        "code_verifier": challenge.verifier,
        "client_id": v1_const.AFERO_CLIENTS["hubspace"]["AUTH_DEFAULT_CLIENT_ID"],
    }
    if expected_messages:
        for expected_message in expected_messages:
            assert expected_message in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "secure_mode,refresh_token,response,expected,expected_message,err",
    [
        # Refresh token invalidated due to password change
        (
            True,
            "code",
            {"status": 400, "body": json.dumps({"error": "invalid_grant"})},
            None,
            None,
            auth.InvalidAuth,
        ),
        # Invalid status
        (True, "code", {"status": 403}, None, None, aiohttp.web_exceptions.HTTPForbidden),
        # Unexpected code returned
        (True, "code", {"status": 400}, None, None, auth.InvalidResponse),
        # bad response
        (
            True,
            "code",
            {"status": 200, "body": json.dumps({"id_token2": "cool_beans"})},
            None,
            None,
            auth.InvalidResponse,
        ),
        # valid response
        (
            True,
            "code",
            {
                "status": 200,
                "body": json.dumps(
                    {
                        "id_token": "cool_beans",
                        "refresh_token": "refresh_beans",
                        "access_token": "access_token_beans",
                    }
                ),
            },
            "refresh_beans",
            (
                "JSON response: {'id_token': 'co***ns', "
                "'refresh_token': 're***ns', 'access_token': 'ac***ns'}"
            ),
            None,
        ),
        # valid response insecure
        (
            False,
            "code",
            {
                "status": 200,
                "body": json.dumps(
                    {
                        "id_token": "cool_beans",
                        "refresh_token": "refresh_beans",
                        "access_token": "access_token_beans",
                    }
                ),
            },
            "refresh_beans",
            (
                "JSON response: {'id_token': 'cool_beans', 'refresh_token': "
                "'refresh_beans', 'access_token': 'access_token_beans'}"
            ),
            None,
        ),
    ],
)
async def test_generate_refresh_token_from_refresh(
    secure_mode,
    refresh_token,
    response,
    expected,
    expected_message,
    err,
    hs_auth,
    aioresponses,
    aio_sess,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    if not secure_mode:
        hs_auth.secret_logger = auth.passthrough
    hs_auth._token_data = auth.TokenData(
        None, None, refresh_token, datetime.datetime.now()
    )
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_TOKEN_ENDPOINT"])
    aioresponses.post(url, **response)
    if expected:
        assert (
            expected == (await hs_auth.generate_refresh_token()).refresh_token
        )
    else:
        with pytest.raises(err):
            await hs_auth.generate_refresh_token()
    aioresponses.assert_called_once()
    call_args = list(aioresponses.requests.values())[0][0]
    # Add in the user-agent that is generated from the bridge
    hs_auth._token_headers["user-agent"] = v1_const.AFERO_GENERICS["DEFAULT_USERAGENT"].safe_substitute(
        client_name="aioafero"
    )
    assert call_args.kwargs["headers"] == hs_auth._token_headers
    assert call_args.kwargs["data"] == {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "scope": "openid email offline_access profile",
        "client_id": v1_const.AFERO_CLIENTS["hubspace"]["AUTH_DEFAULT_CLIENT_ID"],
    }
    if expected_message:
        assert expected_message in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "webapp_login_return, generate_refresh_token_return",
    [
        ("cool", "beans"),
    ],
)
async def test_perform_initial_login(
    webapp_login_return, generate_refresh_token_return, hs_auth, aio_sess, mocker
):
    mocker.patch.object(hs_auth, "webapp_login", return_value=webapp_login_return)
    mocker.patch.object(
        hs_auth, "generate_refresh_token", return_value=generate_refresh_token_return
    )
    assert (
        await hs_auth.perform_initial_login() == generate_refresh_token_return
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "hide_secrets, refresh_token",
    [
        (True, None),
        (False, "yes"),
    ],
)
async def test_AferoAuth_init(hide_secrets, refresh_token, mocker, bridge):
    test_auth = auth.AferoAuth(
        bridge, "username", "password", hide_secrets=hide_secrets, refresh_token=refresh_token
    )
    if hide_secrets:
        assert test_auth.secret_logger == auth.LogRedactorMessage
    else:
        assert test_auth.secret_logger == auth.passthrough
    if refresh_token:
        assert test_auth._token_data == auth.TokenData(
            None, None, refresh_token, mocker.ANY
        )
    else:
        assert test_auth._token_data is None


def bad_refresh_token(*args, **kwargs):
    yield auth.InvalidAuth()
    yield auth.TokenData(
        "token",
        "access_token",
        "refresh_token",
        datetime.datetime.now().timestamp() + 120,
    )


def bad_refresh_token_invalid(*args, **kwargs):
    yield auth.InvalidAuth()
    yield auth.InvalidAuth()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "token_data, results_perform_initial_login, results_generate_refresh_token,expected,messages",
    [
        # Perform full login
        (
            None,
            auth.TokenData(
                "token",
                "access_token",
                "refresh_token",
                datetime.datetime.now().timestamp() + 120,
            ),
            None,
            "token",
            ["Refresh token not present. Generating a new refresh token"],
        ),
        # Previously logged in but expired
        (
            None,
            auth.TokenData(
                "token",
                "access_token",
                "refresh_token",
                datetime.datetime.now().timestamp() - 120,
            ),
            auth.TokenData(
                "token",
                "access_token",
                "refresh_token",
                datetime.datetime.now().timestamp() + 120,
            ),
            "token",
            [
                "Token has not been generated or is expired",
            ],
        ),
        # Invalid refresh token
        (
            None,
            auth.TokenData(
                "token",
                "access_token",
                "refresh_token",
                datetime.datetime.now().timestamp() - 120,
            ),
            bad_refresh_token,
            "token",
            [
                "Token has not been generated or is expired",
                "Provided refresh token is no longer valid.",
                "Refresh token not present. Generating a new refresh token",
            ],
        ),
        # Invalid refresh token and bad login
        (
            None,
            auth.TokenData(
                "token",
                "access_token",
                "refresh_token",
                datetime.datetime.now().timestamp() - 120,
            ),
            bad_refresh_token_invalid,
            None,
            [
                "Token has not been generated or is expired",
                "Provided refresh token is no longer valid.",
                "Refresh token not present. Generating a new refresh token",
            ],
        ),
    ],
)
async def test_token(
    token_data,
    results_perform_initial_login,
    results_generate_refresh_token,
    expected,
    messages,
    caplog,
    mocker,
    bridge,
):
    caplog.set_level(logging.DEBUG)
    test_auth = auth.AferoAuth(bridge, "username", "password")
    test_auth._token_data = token_data
    sess = mocker.Mock()
    mocker.patch.object(
        test_auth,
        "perform_initial_login",
        mocker.AsyncMock(return_value=results_perform_initial_login),
    )
    if isinstance(results_generate_refresh_token, auth.TokenData):
        mocker.patch.object(
            test_auth,
            "generate_refresh_token",
            mocker.AsyncMock(return_value=results_generate_refresh_token),
        )
    elif results_generate_refresh_token:
        mocker.patch.object(
            test_auth,
            "generate_refresh_token",
            side_effect=results_generate_refresh_token(),
        )
    if isinstance(expected, str):
        assert await test_auth.token(sess) == expected
        assert test_auth.refresh_token == "refresh_token"
    else:
        with pytest.raises(auth.InvalidAuth):
            await test_auth.token(sess)
    for message in messages:
        assert message in caplog.text


def test_set_token_data(hs_auth):
    data = auth.TokenData(
        "token",
        "access_token",
        "refresh_token",
        datetime.datetime.now().timestamp() + 120,
    )
    hs_auth.set_token_data(data)
    assert hs_auth._token_data == data


def test_property_refresh_token(bridge):
    _auth = auth.AferoAuth(bridge, "username", "password")
    assert _auth.refresh_token is None
    _auth._token_data = auth.TokenData(
        "token",
        "access_token",
        "refresh_token",
        datetime.datetime.now().timestamp() + 120,
    )
    assert _auth.refresh_token == "refresh_token"


@pytest.mark.parametrize(
    ("page_filename", "expected"),
    [
        # Valid OTP error
        ("auth_webapp_login_otp_failed.html", "Invalid access code."),
        # Can't find OTP error
        ("auth_webapp_login.html", "Unknown error"),
    ]
)
def test_get_kc_error(page_filename, expected):
    with open(os.path.join(current_path, "data", page_filename)) as f:
        page_data = f.read()
    assert auth.get_kc_error(page_data) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("page_filename", "response", "expected_code", "expected_error", "expected_error_match"),
    [
        # Valid OTP submission
        (
            None,
            {
                "status": 302,
                "headers": {
                    "location": (
                        "hubspace-app://loginredirect"
                        "?session_state=sess-state"
                        "&iss=https%3A%2F%2Faccounts.hubspaceconnect.com"
                        "%2Fauth%2Frealms%2Fthd&code=code"
                    )
                },
            },
            "code",
            None,
            None,
        ),
        # Invalid OTP provided
        (
            "auth_webapp_login_otp_failed.html",
            {
                "status": 200,
            },
            None,
            auth.InvalidOTP,
            "Invalid access code.",
        ),
    ]
)
async def test_submit_otp(
    page_filename,
    response,
    expected_code,
    expected_error,
    expected_error_match,
    mock_aioresponse,
    aio_sess,
    hs_auth,
):
    challenge = await hs_auth.generate_challenge_data()
    hs_auth._bridge._web_session = aio_sess
    auth_sess_data = auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id")
    url_params = auth.extract_login_codes(auth_sess_data, hs_auth._afero_client)
    hs_auth._otp_data = {
        "params": url_params,
        "headers": {},
        "challenge": challenge,
    }
    if page_filename:
        with open(os.path.join(current_path, "data", page_filename)) as f:
            response["body"] = f.read()
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
    url = await build_url(url, url_params)
    mock_aioresponse.post(url, **response)
    if expected_code:
        assert (await hs_auth.submit_otp("123456")) == expected_code
    else:
        with pytest.raises(expected_error, match=expected_error_match):
            await hs_auth.submit_otp("123456")


@pytest.mark.asyncio
async def test_perform_otp_login(mock_aioresponse, aio_sess, hs_auth, mocker):
    challenge = await hs_auth.generate_challenge_data()
    hs_auth._bridge._web_session = aio_sess
    auth_sess_data = auth.AuthSessionData("url_sess_code", "url_exec_code", "url_tab_id")
    url_params = auth.extract_login_codes(auth_sess_data, hs_auth._afero_client)
    hs_auth._otp_data = {
        "params": url_params,
        "headers": {},
        "challenge": challenge,
    }
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_CODE_ENDPOINT"])
    url = await build_url(url, url_params)
    # Successful OTP POST
    otp_post_response = {
        "status": 302,
        "headers": {
            "location": (
                "hubspace-app://loginredirect"
                "?session_state=sess-state"
                "&iss=https%3A%2F%2Faccounts.hubspaceconnect.com"
                "%2Fauth%2Frealms%2Fthd&code=code"
            )
        },
    }
    mock_aioresponse.post(url, **otp_post_response)
    # Successful authorization_code generation
    url = hs_auth.generate_auth_url(v1_const.AFERO_GENERICS["AUTH_TOKEN_ENDPOINT"])
    resp_data = {
        "refresh_token": "refresh_token",
        "access_token": "access_token",
        "id_token": "id_token",
    }
    mock_aioresponse.post(url, status=200, body=json.dumps(resp_data))
    assert await hs_auth.perform_otp_login("123456") == auth.TokenData(
        token="id_token",
        access_token="access_token",
        refresh_token="refresh_token",
        expiration=mocker.ANY,
    )

@pytest.mark.asyncio
async def test_perform_otp_login_not_ready(hs_auth):
    hs_auth._otp_data = {}
    with pytest.raises(auth.OTPRequired):
        await hs_auth.perform_otp_login("123456")



"""Controller holding and managing Afero IoT resources of type `security-system-keypad`."""

from aioafero.device import AferoDevice
from aioafero.errors import DeviceNotFound
from aioafero.v1.models import SecuritySystemKeypad, SecuritySystemKeypadPut, features
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes

from .base import AferoBinarySensor, BaseResourcesController


class SecuritySystemKeypadController(BaseResourcesController[SecuritySystemKeypad]):
    """Controller holding and managing Afero IoT resources of type `security-system-keypad`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.SECURITY_SYSTEM_KEYPAD]
    ITEM_CLS = SecuritySystemKeypad
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "tamper-detection": "tampered",
    }
    # Elements that map to Select. func class / func instance to name
    ITEM_SELECTS = {
        ("volume", "buzzer-volume"): "Buzzer Volume",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> SecuritySystemKeypad:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        available: bool = False
        selects: dict[tuple[str, str | None], features.SelectFeature] | None = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        for state in afero_device.states:
            if state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.device_id):
                binary_sensors[sensor.id] = sensor
            elif select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]

        self._items[afero_device.id] = SecuritySystemKeypad(
            _id=afero_device.id,
            available=available,
            binary_sensors=binary_sensors,
            selects=selects,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Security System with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif (
                (update_key := await self.update_sensor(state, cur_item))
                or (update_key := await self.update_number(state, cur_item))
                or (update_key := await self.update_select(state, cur_item))
            ):
                updated_keys.add(update_key)

        return updated_keys

    async def set_state(
        self,
        device_id: str,
        selects: dict[tuple[str, str | None], str] | None = None,
    ) -> None:
        """Set supported feature(s) to Security System resource."""
        update_obj = SecuritySystemKeypadPut()
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if selects:
            for key, val in selects.items():
                if key not in cur_item.selects:
                    continue
                update_obj.selects[key] = features.SelectFeature(
                    selected=val,
                    selects=cur_item.selects[key].selects,
                    name=cur_item.selects[key].name,
                )
        await self.update(device_id, obj_in=update_obj)



import json
import os
from typing import Any

from aioafero.device import AferoDevice, AferoState, AferoCapability

current_path: str = os.path.dirname(os.path.realpath(__file__))


def get_device_dump(file_name: str) -> Any:
    """Get a device dump

    :param file_name: Name of the file to load
    """
    with open(os.path.join(current_path, "device_dumps", file_name)) as fh:
        return json.load(fh)


def get_raw_dump(file_name: str) -> Any:
    """Get a device dump

    :param file_name: Name of the file to load
    """
    with open(os.path.join(current_path, "data", file_name)) as fh:
        return json.load(fh)


def create_devices_from_data(file_name: str) -> list[AferoDevice]:
    """Generate devices from a data dump

    :param file_name: Name of the file to load
    """
    devices = get_device_dump(file_name)
    processed = []
    for device in devices:
        processed.append(create_device_from_data(device))
    return processed


def create_device_from_data(device: dict) -> AferoDevice:
    processed_states = []
    for state in device["states"]:
        processed_states.append(AferoState(**state))
    device["states"] = processed_states
    processed_capabilities = []
    for cap in device.get("capabilities", []):
        processed_capabilities.append(AferoCapability(**cap))
    device["capabilities"] = processed_capabilities
    if "children" not in device:
        device["children"] = []
    return AferoDevice(**device)


def get_json_call(mocked_controller):
    mocked_controller._bridge.request.assert_called_once()
    call = mocked_controller._bridge.request.call_args_list[0][1]
    assert "json" in call
    return call["json"]


def ensure_states_sent(mocked_controller, expected_states, device_id=None):
    req = get_json_call(mocked_controller)
    assert len(req["values"]) == len(
        expected_states
    ), f"States Sent: {len(req)}. Expected: {len(expected_states)}. Actual: {req}"
    for state in expected_states:
        assert state in req["values"], (
            f"Missing {state['functionClass']} / "
            f"{state['functionInstance']} for "
            f"{state['value']} in {req}"
        )
    if device_id:
        assert req["metadeviceId"] == device_id


def modify_state(device: AferoDevice, new_state):
    for ind, state in enumerate(device.states):
        if state.functionClass != new_state.functionClass:
            continue
        if (
            new_state.functionInstance
            and new_state.functionInstance != state.functionInstance
        ):
            continue
        device.states[ind] = new_state
        break


def create_hs_raw_from_device(afero_dev: AferoDevice) -> dict:
    """Convert an AferoDevice to Hubspace data"""
    descr_device = {
        "defaultName": afero_dev.default_name,
        "deviceClass": afero_dev.device_class,
        "manufacturerName": afero_dev.manufacturerName,
        "model": afero_dev.model,
        "profileId": "6ea6d241-3909-4235-836d-c594ece2bb67",
        "type": "device",
    }
    description = {
        "createdTimestampMs": 0,
        "defaultImage": afero_dev.default_image,
        "descriptions": [],
        "device": descr_device,
        "functions": afero_dev.functions,
        "hints": [],
        "id": afero_dev.id,
        "updatedTimestampMs": 0,
        "version": 1,
    }
    return {
        "children": afero_dev.children,
        "createdTimestampMs": 0,
        "description": description,
        "deviceId": afero_dev.device_id,
        "friendlyDescription": "",
        "friendlyName": afero_dev.friendly_name,
        "id": afero_dev.id,
        "state": {
            "metadeviceId": afero_dev.id,
            "values": convert_states(afero_dev.states),
        },
        "typeId": "metadevice.device",
    }


def create_hs_raw_from_dump(file_name: str) -> list[dict]:
    """Generate a Hubspace payload from devices and save it to a file.

    Takes a device dump file, processes it into Hubspace format, and saves the
    result to a new JSON file with '-raw' suffix. The generated payload includes
    device details, descriptions, states and other metadata formatted for Hubspace.

    :param file_name: Name of the file that contains the dump
    :return: List of dictionaries containing the generated Hubspace payload
    """
    hs_raw: list[dict] = []
    for device in create_devices_from_data(file_name):
        hs_raw.append(create_hs_raw_from_device(device))
    return hs_raw


def convert_states(states: list[AferoState]) -> list[dict]:
    """Convert the states from AferoState to raw.

    :param states: List of AferoState objects
    """
    return [
        {
            "functionClass": state.functionClass,
            "functionInstance": state.functionInstance,
            "lastUpdateTime": state.lastUpdateTime,
            "value": state.value,
        }
        for state in states
    ]



import datetime
import asyncio

from aioresponses import aioresponses
import pytest
import pytest_asyncio
import aiohttp

from aioafero.v1 import AferoBridgeV1
from aioafero import AferoDevice
from aioafero.v1.auth import TokenData
from aioafero.v1.controllers.event import EventType
from tests.v1.utils import create_hs_raw_from_device
from aioafero.v1.controllers.base import BaseResourcesController, dataclass_to_afero
import securelogging


@pytest.fixture(autouse=True)
def reset_logging_secrets():
    securelogging._called_from_test = True
    securelogging.reset_secrets()
    yield


@pytest_asyncio.fixture(scope="function")
async def aio_sess() -> aiohttp.ClientSession:
    yield aiohttp.ClientSession()


@pytest_asyncio.fixture
async def mocked_bridge(mocker, aio_sess) -> AferoBridgeV1:
    """Create a mocked afero bridge to be used in tests."""
    mocker.patch("time.time", return_value=12345)
    mocker.patch("aioafero.v1.controllers.event.EventStream.gather_data")

    bridge: AferoBridgeV1 = AferoBridgeV1("username2", "password2")
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "fetch_data", return_value=[])
    mocker.patch.object(bridge.events, "initialize_reader")
    mocker.patch.object(bridge, "request", side_effect=mocker.AsyncMock())
    mocker.patch.object(
        bridge, "fetch_data", side_effect=mocker.AsyncMock(return_value=[])
    )
    mocker.patch.object(bridge.events, "_first_poll_completed", True)
    mocker.patch.object(bridge, "_web_session", aio_sess)

    bridge.set_token_data(
        TokenData(
            "mock-token",
            "mock-access",
            "mock-refresh-token",
            expiration=datetime.datetime.now().timestamp() + 200,
        )
    )

    # Enable ad-hoc polls
    async def generate_events_from_data(data):
        task = asyncio.create_task(bridge.events.generate_events_from_data(data))
        await task
        raw_data = await bridge.events.generate_events_from_data(data)
        mocker.patch(
            "aioafero.v1.controllers.event.EventStream.gather_data",
            return_value=raw_data,
        )
        await bridge.async_block_until_done()

    # Fake a poll for discovery
    async def generate_devices_from_data(devices: list[AferoDevice]):
        raw_data = [create_hs_raw_from_device(device) for device in devices]
        mocker.patch(
            "aioafero.v1.controllers.event.EventStream.gather_data",
            return_value=raw_data,
        )
        await bridge.events.generate_events_from_data(raw_data)
        await bridge.async_block_until_done()

    # Fake the response from the API when updating states
    def mock_update_afero_api(device_id, result):
        json_resp = mocker.AsyncMock()
        json_resp.return_value = {"metadeviceId": device_id, "values": result}
        resp = mocker.AsyncMock()
        resp.json = json_resp
        resp.status = 200
        mocker.patch("aioafero.v1.controllers.base.BaseResourcesController.update_afero_api", return_value=resp)
    
    # Enable "results" to be returned on update
    actual_dataclass_to_afero = dataclass_to_afero
    def mocked_dataclass_to_afero(*args, **kwargs):
        result = actual_dataclass_to_afero(*args, **kwargs)
        mock_update_afero_api(args[0].id, result)
        return result
    
    mocker.patch("aioafero.v1.controllers.base.dataclass_to_afero", side_effect=mocked_dataclass_to_afero)
    
    bridge.mock_update_afero_api = mock_update_afero_api
    bridge.generate_devices_from_data = generate_devices_from_data
    bridge.generate_events_from_data = generate_events_from_data

    await bridge.initialize()
    yield bridge
    await bridge.close()


@pytest.fixture
def mocked_bridge_req(mocker, aio_sess):
    bridge: AferoBridgeV1 = AferoBridgeV1("username2", "password2")
    mocker.patch.object(
        bridge,
        "get_account_id",
        side_effect=mocker.AsyncMock(return_value="mocked-account-id"),
    )
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "initialize", side_effect=mocker.AsyncMock())
    mocker.patch.object(bridge, "fetch_data", side_effect=bridge.fetch_data)
    mocker.patch.object(bridge, "request", side_effect=bridge.request)
    mocker.patch.object(bridge, "_web_session", aio_sess)
    mocker.patch.object(bridge.events, "_first_poll_completed", True)
    bridge._auth._token_data = TokenData(
        "mock-token",
        None,
        "mock-refresh-token",
        expiration=datetime.datetime.now().timestamp() + 200,
    )
    # Force initialization so test elements are not overwritten
    for controller in bridge._controllers.values():
        controller._initialized = True

    # Enable ad-hoc event updates
    def emit_event(event_type, data):
        bridge.events.emit(EventType(event_type), data)

    # Enable ad-hoc polls
    async def generate_events_from_data(data):
        task = asyncio.create_task(bridge.events.generate_events_from_data(data))
        await task
        raw_data = await bridge.events.generate_events_from_data(data)
        mocker.patch(
            "aioafero.v1.controllers.event.EventStream.gather_data",
            return_value=raw_data,
        )
        await bridge.async_block_until_done()

    # Fake a poll for discovery
    async def generate_devices_from_data(devices: list[AferoDevice]):
        raw_data = [create_hs_raw_from_device(device) for device in devices]
        mocker.patch(
            "aioafero.v1.controllers.event.EventStream.gather_data",
            return_value=raw_data,
        )
        await bridge.events.generate_events_from_data(raw_data)
        await bridge.async_block_until_done()

    bridge.emit_event = emit_event
    bridge.generate_devices_from_data = generate_devices_from_data
    bridge.generate_events_from_data = generate_events_from_data

    bridge.__aenter__ = mocker.AsyncMock(return_value=bridge)
    bridge.__aexit__ = mocker.AsyncMock()
    return bridge


@pytest_asyncio.fixture
async def bridge(mocker):
    bridge = AferoBridgeV1("user", "passwd")
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "fetch_data", return_value=[])
    mocker.patch.object(bridge, "request", side_effect=mocker.AsyncMock())
    mocker.patch.object(bridge.events, "_first_poll_completed", True)
    await bridge.initialize()
    await bridge.async_block_until_done()
    yield bridge
    await bridge.close()


@pytest_asyncio.fixture
async def bridge_with_acct(mocker):
    bridge = AferoBridgeV1("user", "passwd")
    bridge._auth._token_data = TokenData(
            "mock-token",
            None,
            "mock-refresh-token",
            expiration=datetime.datetime.now().timestamp() + 200,
        )
    yield bridge


@pytest_asyncio.fixture
async def bridge_with_acct_req(mocker):
    bridge = AferoBridgeV1("user", "passwd")
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "request", side_effect=bridge.request)
    mocker.patch.object(bridge.events, "_first_poll_completed", True)
    bridge._auth._token_data = TokenData(
            "mock-token",
            None,
            "mock-refresh-token",
            expiration=datetime.datetime.now().timestamp() + 200,
        )
    await bridge.initialize()
    await bridge.async_block_until_done()
    yield bridge
    await bridge.close()


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m



"""Test ExhaustFanController"""

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.exhaust_fan import (
    ExhaustFanController,
    exhaust_fan_callback,
    features,
    generate_split_name,
    get_split_instances,
    get_valid_states,
)
from aioafero.v1.models import ResourceTypes
from aioafero.v1.models.sensor import AferoBinarySensor

from .. import utils

exhaust_fan = utils.create_devices_from_data("exhaust-fan.json")[0]
a21_light = utils.create_devices_from_data("light-a21.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.exhaust_fans


def test_generate_split_name():
    assert (
        generate_split_name(exhaust_fan, "humidity-detection-enabled")
        == "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-humidity-detection-enabled"
    )


def test_get_split_instances():
    assert get_split_instances(exhaust_fan) == [
        "humidity-detection-enabled",
        "humidity-sensor-led",
        "motion-detection-enabled-exhaust-fan",
        "motion-sensor-led",
        "speaker-power",
    ]


def test_get_valid_states():
    assert get_valid_states(exhaust_fan, "humidity-detection-enabled") == [
        AferoState(
            functionClass="toggle",
            value="off",
            lastUpdateTime=0,
            functionInstance="humidity-detection-enabled",
        ),
        AferoState(
            functionClass="available",
            value=True,
            lastUpdateTime=0,
            functionInstance=None,
        ),
    ]


def test_exhaust_fan_callback():
    devs, remove_original = exhaust_fan_callback(exhaust_fan)
    assert remove_original is False
    assert len(devs) == 5
    expected_ids = [
        "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-humidity-detection-enabled",
        "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-humidity-sensor-led",
        "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-motion-detection-enabled-exhaust-fan",
        "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-motion-sensor-led",
        "44620d02-8b62-49ce-afe8-1ea8f15e0ec5-exhaust-fan-speaker-power",
    ]
    for ind, dev in enumerate(devs):
        assert dev.id == expected_ids[ind]
        assert len(dev.states) == 2
        assert dev.device_class == ResourceTypes.SWITCH.value


def test_exhaust_fan_callback_wrong_dev():
    devs, remove_original = exhaust_fan_callback(a21_light)
    assert remove_original is False
    assert len(devs) == 0


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "44620d02-8b62-49ce-afe8-1ea8f15e0ec5"
    assert dev.binary_sensors == {
        "humidity-threshold-met|humidity-threshold-met": AferoBinarySensor(
            id="humidity-threshold-met|humidity-threshold-met",
            owner="7475607b-c3ea-4afe-a465-1c11c5c09985",
            current_value="below-threshold",
            _error="above-threshold",
            unit=None,
            instance="humidity-threshold-met",
        ),
        "motion-detection|motion-detection": AferoBinarySensor(
            id="motion-detection|motion-detection",
            owner="7475607b-c3ea-4afe-a465-1c11c5c09985",
            current_value="motion-detected",
            _error="motion-detected",
            unit=None,
            instance="motion-detection",
        ),
    }
    assert dev.sensors == {}
    assert dev.numbers == {
        ("auto-off-timer", "auto-off"): features.NumbersFeature(
            value=1200,
            min=60,
            max=1800,
            step=60,
            name="Auto Off Timer",
            unit="seconds",
        )
    }
    assert dev.selects == {
        ("motion-action", "exhaust-fan"): features.SelectFeature(
            selected="light", selects={"both", "fan", "light"}, name="Motion Action"
        ),
        ("sensitivity", "humidity-sensitivity"): features.SelectFeature(
            selected="3-medium",
            selects={"1-low", "2-low-medium", "3-medium", "4-medium-high", "5-high"},
            name="Humidity Sensitivity",
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(exhaust_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    new_states = [
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="humidity-threshold-met", value="above-threshold", lastUpdateTime=0, functionInstance="humidity-threshold-met"
        ),
        AferoState(
            functionClass="auto-off-timer", value=120, lastUpdateTime=0, functionInstance="auto-off"
        ),
        AferoState(
            functionClass="motion-action", value="both", lastUpdateTime=0, functionInstance="exhaust-fan"
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is False
    assert (
        dev.binary_sensors["humidity-threshold-met|humidity-threshold-met"].current_value
        == "above-threshold"
    )
    assert dev.numbers[("auto-off-timer", "auto-off")].value == 120
    assert dev.selects[("motion-action", "exhaust-fan")].selected == "both"
    assert updates == {
        "binary-humidity-threshold-met|humidity-threshold-met",
        "available",
        "number-('auto-off-timer', 'auto-off')",
        "select-('motion-action', 'exhaust-fan')",
    }


@pytest.mark.asyncio
async def test_update_state_no_change(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(exhaust_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    new_states = [
        AferoState(
            functionClass="available", value=True, lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is True
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(exhaust_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.set_state(exhaust_fan.id)


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(exhaust_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    await mocked_controller.set_state(
        exhaust_fan.id,
        numbers={
            ("auto-off-timer", "auto-off"): 120,
            ("invalid", "state"): None,
        },
        selects={
            ("motion-action", "exhaust-fan"): "both",
            ("invalid", "state"): None,
        },
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller.items[0]
    assert dev.numbers[("auto-off-timer", "auto-off")].value == 120
    assert dev.selects[("motion-action", "exhaust-fan")].selected == "both"


@pytest.mark.asyncio
async def test_exhaust_fan_emitting(bridge):
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    await bridge.async_block_until_done()
    assert len(bridge.exhaust_fans._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    await bridge.async_block_until_done()
    assert len(bridge.exhaust_fans._items) == 1
    assert not bridge.exhaust_fans._items[dev_update.id].available


@pytest.mark.asyncio
async def test_set_state_no_dev(mocked_controller, caplog):
    caplog.set_level(0)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(exhaust_fan)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller._bridge.add_device(exhaust_fan.id, mocked_controller)
    await mocked_controller.set_state("not-a-device")
    mocked_controller._bridge.request.assert_not_called()
    assert "Unable to find device" in caplog.text



from aioafero.v1.models import features


def test_ColorModeFeature():
    feat = features.ColorModeFeature("white")
    assert feat.api_value == "white"


def test_ColorFeature():
    feat = features.ColorFeature(red=10, green=20, blue=30)
    assert feat.api_value == {
        "value": {
            "color-rgb": {
                "r": 10,
                "g": 20,
                "b": 30,
            }
        }
    }


def test_ColorTemperatureFeature():
    feat = features.ColorTemperatureFeature(
        temperature=3000, supported=[1000, 2000, 3000], prefix="K"
    )
    assert feat.api_value == "3000K"


def test_CurrentPositionEnum():
    feat = features.CurrentPositionEnum("locking")
    assert feat.value == features.CurrentPositionEnum.LOCKING.value
    feat = features.CurrentPositionEnum("no")
    assert feat.value == features.CurrentPositionEnum.UNKNOWN.value


def test_CurrentPositionFeature():
    feat = features.CurrentPositionFeature(features.CurrentPositionEnum.LOCKED)
    assert feat.api_value == "locked"


def test_CurrentTemperatureFeature():
    feat = features.CurrentTemperatureFeature(
        temperature=1, function_class="temperature", function_instance="current-temp"
    )
    assert feat.api_value == {
        "functionClass": "temperature",
        "functionInstance": "current-temp",
        "value": 1,
    }


def test_DimmingFeature():
    feat = features.DimmingFeature(
        brightness=30, supported=[10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    )
    assert feat.api_value == 30


def test_DirectionFeature():
    feat = features.DirectionFeature(forward=True)
    assert feat.api_value == "forward"
    feat = features.DirectionFeature(forward=False)
    assert feat.api_value == "reverse"


def test_EffectFeature():
    feat = features.EffectFeature(
        effect="fade-3", effects={"preset": {"fade-3"}, "custom": {"rainbow"}}
    )
    assert feat.api_value == [
        {
            "functionClass": "color-sequence",
            "functionInstance": "preset",
            "value": "fade-3",
        }
    ]
    feat.effect = "rainbow"
    assert feat.api_value == [
        {
            "functionClass": "color-sequence",
            "functionInstance": "preset",
            "value": "custom",
        },
        {
            "functionClass": "color-sequence",
            "functionInstance": "custom",
            "value": "rainbow",
        },
    ]
    assert feat.is_preset("fade-3")
    assert not feat.is_preset("rainbow")
    # effect does not exist
    feat.effect = "nope"
    assert feat.api_value == []
    feat = features.EffectFeature(effect="fade-3", effects={"custom": {"rainbow"}})
    assert not feat.is_preset("rainbow")
    


def test_HVACModeFeature():
    feat = features.HVACModeFeature(
        mode="beans",
        previous_mode="not_beans",
        modes={"beans", "not_beans"},
        supported_modes={"beans", "not_beans"},
    )
    assert feat.api_value == "beans"


def test_ModeFeature():
    feat = features.ModeFeature(mode="color", modes={"color", "white"})
    assert feat.api_value == "color"


def test_NumbersFeature():
    feat = features.NumbersFeature(
        value=12,
        min=0,
        max=20,
        step=1,
        name="Cool Beans",
        unit="bean count",
    )
    assert feat.api_value == 12


def test_OnFeature():
    feat = features.OnFeature(on=True)
    assert feat.api_value == {"value": "on", "functionClass": "power"}
    feat = features.OnFeature(on=False, func_class="cool", func_instance="beans")
    assert feat.api_value == {
        "value": "off",
        "functionClass": "cool",
        "functionInstance": "beans",
    }


def test_OpenFeature():
    feat = features.OpenFeature(open=True)
    assert feat.api_value == {"value": "on", "functionClass": "toggle"}
    feat = features.OpenFeature(open=False, func_class="cool", func_instance="beans")
    assert feat.api_value == {
        "value": "off",
        "functionClass": "cool",
        "functionInstance": "beans",
    }


def test_PresetFeature():
    feat = features.PresetFeature(
        enabled=True, func_class="cool", func_instance="beans"
    )
    assert feat.api_value == {
        "value": "enabled",
        "functionClass": "cool",
        "functionInstance": "beans",
    }
    feat.enabled = False
    assert feat.api_value == {
        "value": "disabled",
        "functionClass": "cool",
        "functionInstance": "beans",
    }


def test_SelectFeature():
    feat = features.SelectFeature(
        selected="beans", selects={"cool", "beans"}, name="Those beans"
    )
    assert feat.api_value == "beans"


def test_SpeedFeature():
    feat = features.SpeedFeature(
        speed=25,
        speeds=[
            "speed-4-0",
            "speed-4-25",
            "speed-4-50",
            "speed-4-75",
            "speed-4-100",
        ],
    )
    assert feat.api_value == "speed-4-25"
    feat.speed = 50
    assert feat.api_value == "speed-4-50"


def test_TargetTemperatureAutoFeature():
    feat = features.TargetTemperatureFeature(
        value=12,
        min=10,
        max=14,
        step=0.5,
        instance="whatever",
    )
    assert feat.api_value == {
        "functionClass": "temperature",
        "functionInstance": "whatever",
        "value": 12,
    }


def test_SecuritySensorSirenFeature():
    feat = features.SecuritySensorSirenFeature(
        result_code=0,
        command=4,
    )
    assert feat.api_value == {
        "functionClass": "siren-action",
        "value": {"security-siren-action": {"resultCode": 0, "command": 4}},
        "functionInstance": None,
    }
    feat = features.SecuritySensorSirenFeature(
        result_code=None,
        command=None,
    )
    assert feat.api_value == {
        "functionClass": "siren-action",
        "value": None,
        "functionInstance": None,
    }



"""Controller holding and managing Afero IoT resources of type `thermostat`."""

import copy

from aioafero.device import AferoDevice
from aioafero.util import process_function
from aioafero.v1.models import features
from aioafero.v1.models.resource import DeviceInformation, ResourceTypes
from aioafero.v1.models.thermostat import Thermostat, ThermostatPut

from .climate import ClimateController


class ThermostatController(ClimateController[Thermostat]):
    """Controller holding and managing Afero IoT resources of type `thermostat`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.THERMOSTAT]
    ITEM_CLS = Thermostat
    ITEM_MAPPING = {
        "fan_mode": "fan-mode",
        "hvac_mode": "mode",
    }
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "filter-replacement": "replacement-needed",
        "max-temp-exceeded": "alerting",
        "min-temp-exceeded": "alerting",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> Thermostat:
        """Initialize the element.

        :param afero_device: Afero Device that contains the updated states

        :return: Newly initialized resource
        """
        climate_data = await self.initialize_climate_elem(afero_device)

        fan_running: bool | None = None
        fan_mode: features.ModeFeature | None = None
        hvac_action: str | None = None
        system_type: str | None = None

        for state in afero_device.states:
            if state.functionClass == "fan-mode":
                fan_mode = features.ModeFeature(
                    mode=state.value,
                    modes=set(process_function(afero_device.functions, "fan-mode")),
                )
            elif state.functionClass == "current-fan-state":
                fan_running = state.value == "on"
            elif state.functionClass == "current-system-state":
                hvac_action = state.value
            elif state.functionClass == "system-type":
                system_type = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.id):
                climate_data["binary_sensors"][sensor.id] = sensor

        # Determine supported modes
        climate_data["hvac_mode"].supported_modes = get_supported_modes(
            system_type, climate_data["hvac_mode"].modes
        )

        self._items[afero_device.id] = Thermostat(
            _id=afero_device.id,
            available=climate_data["available"],
            sensors=climate_data["sensors"],
            binary_sensors=climate_data["binary_sensors"],
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
                children=afero_device.children,
                functions=afero_device.functions,
            ),
            current_temperature=climate_data["current_temperature"],
            fan_running=fan_running,
            fan_mode=fan_mode,
            hvac_action=hvac_action,
            hvac_mode=climate_data["hvac_mode"],
            safety_max_temp=climate_data["safety_max_temp"],
            safety_min_temp=climate_data["safety_min_temp"],
            target_temperature_auto_cooling=climate_data[
                "target_temperature_auto_cooling"
            ],
            target_temperature_auto_heating=climate_data[
                "target_temperature_auto_heating"
            ],
            target_temperature_cooling=climate_data["target_temperature_cooling"],
            target_temperature_heating=climate_data["target_temperature_heating"],
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        """Update the Thermostat with the latest API data.

        :param afero_device: Afero Device that contains the updated states

        :return: States that have been modified
        """
        updated_keys = await self.update_climate_elem(afero_device)
        cur_item = self.get_device(afero_device.id)

        for state in afero_device.states:
            if state.functionClass == "current-fan-state":
                temp_val = state.value == "on"
                if cur_item.fan_running != temp_val:
                    cur_item.fan_running = temp_val
                    updated_keys.add("current-fan-state")
            elif state.functionClass == "fan-mode":
                if cur_item.fan_mode.mode != state.value:
                    cur_item.fan_mode.mode = state.value
                    updated_keys.add("fan-mode")
            elif state.functionClass == "current-system-state":
                if cur_item.hvac_action != state.value:
                    cur_item.hvac_action = state.value
                    updated_keys.add(state.functionClass)
        return updated_keys

    async def set_fan_mode(self, device_id: str, fan_mode: str) -> None:
        """Enable or disable fan mode."""
        return await self.set_state(device_id, fan_mode=fan_mode)

    async def set_hvac_mode(self, device_id: str, hvac_mode: str) -> None:
        """Set the current mode of the HVAC system."""
        return await self.set_state(device_id, hvac_mode=hvac_mode)

    async def set_target_temperature(
        self, device_id: str, target_temperature: float
    ) -> None:
        """Set the target temperature."""
        return await self.set_state(device_id, target_temperature=target_temperature)

    async def set_temperature_range(
        self, device_id: str, temp_low: float, temp_high: float
    ) -> None:
        """Set the temperature range for the thermostat."""
        return await self.set_state(
            device_id,
            target_temperature_auto_heating=temp_low,
            target_temperature_auto_cooling=temp_high,
        )

    async def set_state(
        self,
        device_id: str,
        fan_mode: str | None = None,
        hvac_mode: str | None = None,
        safety_max_temp: float | None = None,
        safety_min_temp: float | None = None,
        target_temperature_auto_heating: float | None = None,
        target_temperature_auto_cooling: float | None = None,
        target_temperature_heating: float | None = None,
        target_temperature_cooling: float | None = None,
        **kwargs,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = ThermostatPut()
        cur_item = self.get_device(device_id)
        if fan_mode is not None:
            if fan_mode in cur_item.fan_mode.modes:
                update_obj.fan_mode = features.ModeFeature(
                    mode=fan_mode,
                    modes=cur_item.fan_mode.modes,
                )
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode="fan",
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown fan mode %s. Available modes: %s",
                    fan_mode,
                    ", ".join(sorted(cur_item.fan_mode.modes)),
                )
        if hvac_mode is not None and not update_obj.hvac_mode:
            if hvac_mode in cur_item.hvac_mode.supported_modes:
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode=hvac_mode,
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown hvac mode %s. Available modes: %s",
                    hvac_mode,
                    ", ".join(sorted(cur_item.hvac_mode.supported_modes)),
                )
        # Setting the temp without a specific means we need to adjust the active
        # mode.
        target_temperature = kwargs.pop("target_temperature", None)
        if target_temperature:
            if hvac_mode and hvac_mode in cur_item.hvac_mode.supported_modes:
                mode_to_set = hvac_mode
            else:
                mode_to_set = cur_item.get_mode_to_check()
            if mode_to_set == "cool":
                target_temperature_cooling = target_temperature
                kwargs["target_temperature_cooling"] = target_temperature
            elif mode_to_set == "heat":
                target_temperature_heating = target_temperature
                kwargs["target_temperature_heating"] = target_temperature
            else:
                self._logger.debug(
                    "Unable to set the target temperature due to the active mode: %s",
                    cur_item.hvac_mode.mode,
                )
        kwargs["safety_min_temp"] = safety_min_temp
        kwargs["safety_max_temp"] = safety_max_temp
        kwargs["target_temperature_auto_heating"] = target_temperature_auto_heating
        kwargs["target_temperature_auto_cooling"] = target_temperature_auto_cooling
        kwargs["target_temperature_heating"] = target_temperature_heating
        kwargs["target_temperature_cooling"] = target_temperature_cooling
        await self.set_climate_state(device_id, update_obj, **kwargs)


def get_supported_modes(system_type: str, all_modes: set[str]) -> set:
    """Determine the supported modes based on the system_type."""
    supported_modes = copy.copy(all_modes)
    if "heat-pump" in system_type:
        supports_heating = True
        supports_cooling = True
    else:
        supports_heating = "heating" in system_type
        supports_cooling = "cooling" in system_type
    if not supports_heating and "heat" in supported_modes:
        supported_modes.remove("heat")
    if not supports_cooling and "cool" in supported_modes:
        supported_modes.remove("cool")
    if (not supports_cooling or not supports_heating) and "auto" in supported_modes:
        supported_modes.remove("auto")
    return supported_modes



"""Representation of an Afero Lock and its corresponding updates."""

from dataclasses import dataclass

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class Lock(StandardMixin):
    """Representation of an Afero Lock."""

    type: ResourceTypes = ResourceTypes.LOCK

    position: features.CurrentPositionFeature | None = None


@dataclass
class LockPut:
    """States that can be updated for a Lock."""

    position: features.CurrentPositionFeature | None = None



"""Handle connecting to Afero IoT and distribute events."""

import asyncio
from asyncio.coroutines import iscoroutinefunction
from collections.abc import Callable
import contextlib
import datetime
from enum import Enum
from types import NoneType
from typing import TYPE_CHECKING, Any, NamedTuple, NotRequired, TypedDict

from aiohttp.client_exceptions import ClientError
from aiohttp.web_exceptions import HTTPForbidden, HTTPTooManyRequests

from aioafero.device import AferoDevice, get_afero_device
from aioafero.errors import InvalidAuth
from aioafero.types import EventType
from aioafero.v1.models import ResourceTypes
from aioafero.v1.v1_const import VERSION_POLL_INTERVAL_SECONDS

if TYPE_CHECKING:  # pragma: no cover
    from aioafero.v1 import AferoBridgeV1


class BackoffException(Exception):
    """Exception raised when a backoff is required."""


class CallbackResponse(NamedTuple):
    """Callback response for DEVICE_SPLIT_CALLBACKS.

    :param split_devices: New devices that should be added to the overall list
    :param remove_original: Remove the original device from the list of devices
    """

    split_devices: list[AferoDevice] = []
    remove_original: bool = False


class EventStreamStatus(Enum):
    """Status options of EventStream."""

    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class AferoEvent(TypedDict):
    """Afero IoT Event message as emitted by the EventStream."""

    type: EventType  # = EventType (add, update, delete)
    device_id: NotRequired[str]  # ID for interacting with the device
    device: NotRequired[AferoDevice]  # Afero Device
    polled_data: NotRequired[Any]  # All data polled from the API
    polled_devices: NotRequired[Any]  # All devices after the device split callbacks
    force_forward: NotRequired[bool]


EventCallBackType = Callable[[EventType, dict | None], None]
EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
    "tuple[ResourceTypes] | None",
]


class EventStream:
    """Data gatherer and eventer.

    Polls Afero IoT API, converts the response into devices, and notifies subscribers
    of the event.
    """

    def __init__(
        self, bridge: "AferoBridgeV1", polling_interval: int, poll_version: bool
    ) -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._listeners = set()
        self._event_queue = asyncio.Queue()
        self._status = EventStreamStatus.DISCONNECTED
        self._scheduled_tasks: list[asyncio.Task] = []
        self._subscribers: list[EventSubscriptionType] = []
        self._logger = bridge.logger.getChild("events")
        self._polling_interval: int = polling_interval
        self._multiple_device_finder: dict[str, callable] = {}
        self._version_poll_time: datetime.datetime | None = None
        self._version_poll_enabled: bool = poll_version
        self._first_poll_completed: bool = False

    @property
    def connected(self) -> bool:
        """Return bool if we're connected."""
        return self._status == EventStreamStatus.CONNECTED

    @property
    def status(self) -> EventStreamStatus:
        """Return connection status."""
        return self._status

    @property
    def registered_multiple_devices(self) -> dict[str, Callable]:
        """Get all registered callbacks for splitting devices."""
        return self._multiple_device_finder

    @property
    def polling_interval(self) -> int:
        """Number of seconds between polling."""
        return self._polling_interval

    @polling_interval.setter
    def polling_interval(self, polling_interval: int) -> None:
        """Set the time between polling Afero API."""
        self._polling_interval = polling_interval

    @property
    def poll_version(self) -> bool:
        """Determine if version polling should occur."""
        if not self._version_poll_enabled:
            return False
        now = datetime.datetime.now(datetime.UTC)
        if self._version_poll_time is None:
            self._version_poll_time = now
            return True
        if (
            now - self._version_poll_time
        ).total_seconds() >= VERSION_POLL_INTERVAL_SECONDS:
            self._version_poll_time = now
            return True
        return False

    async def wait_for_first_poll(self) -> None:
        """Wait until the first poll has completed."""
        while not self._first_poll_completed:
            await asyncio.sleep(0.05)

    async def initialize(self) -> None:
        """Start the polling processes."""
        if len(self._scheduled_tasks) == 0:
            await self.initialize_reader()
            await self.initialize_processor()

    async def initialize_reader(self) -> None:
        """Initialize gathering data from Afero API."""
        self._scheduled_tasks.append(asyncio.create_task(self.__event_reader()))

    async def initialize_processor(self) -> None:
        """Initialize the processor."""
        self._scheduled_tasks.append(asyncio.create_task(self.__event_processor()))

    def register_multi_device(self, name: str, generate_devices: callable):
        """Register a callable to find multi-devices within the payload.

        The callable must return a list of tracked AferoDevices
        """
        self._multiple_device_finder[name] = generate_devices

    async def stop(self) -> None:
        """Stop listening for events."""
        with contextlib.suppress(asyncio.CancelledError):
            for task in self._scheduled_tasks:
                task.cancel()
            self._status = EventStreamStatus.DISCONNECTED
            self._scheduled_tasks = []

    def subscribe(
        self,
        callback: EventCallBackType,
        event_filter: EventType | tuple[EventType] | None = None,
        resource_filter: tuple[str] | None = None,
    ) -> Callable:
        """Subscribe to events emitted.

        :param callback: callback function to call when an event emits.
        :param event_filter:  Optionally provide an EventType as filter.
        :param resource_filter: Optionally provide a ResourceType as filter.

        Returns:
            function to unsubscribe.

        """
        if not isinstance(event_filter, NoneType | tuple):
            event_filter = (event_filter,)
        if not isinstance(resource_filter, NoneType | tuple):
            resource_filter = (resource_filter,)
        subscription = (callback, event_filter, resource_filter)

        def unsubscribe():
            self._subscribers.remove(subscription)

        self._subscribers.append(subscription)
        return unsubscribe

    def add_job(self, event: AferoEvent) -> None:
        """Manually add a job to be processed."""
        self._event_queue.put_nowait(event)

    async def async_block_until_done(self):
        """Blocking call until everything has finished."""
        attempt = 0
        while not self._event_queue.empty():
            self._logger.debug(
                "Number of events in queue: %d", self._event_queue.qsize()
            )
            await asyncio.sleep(0.01)
            attempt += 1
            if attempt > 100:
                self._logger.warning(
                    "Queue did not empty within a second. Breaking out of the wait"
                )
                break

    def emit(self, event_type: EventType, data: AferoEvent = None) -> None:
        """Emit event to all listeners."""
        for callback, event_filter, resource_filter in self._subscribers:
            try:
                if event_filter is not None and event_type not in event_filter:
                    continue
                if (
                    resource_filter is not None
                    and data is not None
                    and (
                        "device" in data
                        and data["device"]
                        and not any(
                            data["device"].device_class == res_filter
                            for res_filter in resource_filter
                        )
                    )
                ):
                    continue
                if iscoroutinefunction(callback):
                    self._bridge.add_job(
                        asyncio.create_task(callback(event_type, data))
                    )
                else:
                    callback(event_type, data)
            except Exception:
                self._logger.exception("Unhandled exception. Please open a bug report")

    async def process_backoff(self, attempt: int) -> None:
        """Handle backoff timer for Afero IoT API.

        :param attempt: Number of attempts
        :param reason: Reason why the backoff is occurring
        """
        backoff_time = min(attempt * self.polling_interval, 600)
        debug_message = f"Waiting {backoff_time} seconds before next poll"
        if attempt == 1:
            self._logger.info("Lost connection to the Afero IoT API.")
            self._logger.debug(debug_message)
        if self._status != EventStreamStatus.DISCONNECTED:
            self._status = EventStreamStatus.DISCONNECTED
            self.emit(EventType.DISCONNECTED)
        await asyncio.sleep(backoff_time)

    async def gather_data(self) -> list[dict[Any, str]]:
        """Gather all data from the Afero IoT API."""
        consecutive_http_errors = 0
        while True:
            try:
                data = await self._bridge.fetch_data(version_poll=self.poll_version)
            except TimeoutError:
                self._logger.warning("Timeout when contacting Afero IoT API.")
                await self.process_backoff(consecutive_http_errors)
            except InvalidAuth:
                consecutive_http_errors += 1
                self._logger.warning("Invalid credentials provided.")
                await self.process_backoff(consecutive_http_errors)
            except (HTTPForbidden, HTTPTooManyRequests, ClientError):
                consecutive_http_errors += 1
                await self.process_backoff(consecutive_http_errors)
            except TypeError as err:
                self._logger.warning(
                    "Unexpected data from Afero IoT API, %s.", err.args[0]
                )
                consecutive_http_errors += 1
                await self.process_backoff(consecutive_http_errors)
            except Exception:
                self._logger.exception(
                    "Unknown error occurred. Please open a bug report."
                )
                raise
            else:
                # Successful connection
                if consecutive_http_errors > 0:
                    self._logger.info("Reconnected to the Afero IoT API")
                    self.emit(EventType.RECONNECTED)
                elif self._status != EventStreamStatus.CONNECTED:
                    self._status = EventStreamStatus.CONNECTED
                    self.emit(EventType.CONNECTED)
                return data

    async def generate_devices_from_data(
        self, data: list[dict[Any, str]]
    ) -> list[AferoDevice]:
        """Generate all devices from a given payload.

        Generating devices will attempt to split devices where required and remove
        devices that are no longer needed, as identified by the callback.
        """
        devices = [
            get_afero_device(dev)
            for dev in data
            if dev.get("typeId") == ResourceTypes.DEVICE.value
            and dev.get("description", {}).get("device", {}).get("deviceClass")
        ]
        for device in devices:
            self._bridge.add_afero_dev(device)
        return await self.split_devices(devices)

    async def split_devices(self, devices: list[AferoDevice]) -> list[AferoDevice]:
        """Split Afero devices into multiple devices where required."""
        for name, multi_dev_callable in self._multiple_device_finder.items():
            for dev in devices[:]:
                multi_devs, remove_root = multi_dev_callable(dev)
                if remove_root:
                    with contextlib.suppress(KeyError):
                        devices.remove(dev)
                if multi_devs:
                    self._logger.debug(
                        "Found %s devices from %s", len(multi_devs), name
                    )
                    for split_dev in multi_devs:
                        self._bridge.add_afero_dev(dev, split_dev.id)
                        dev.children.append(split_dev.id)
                    devices.extend(multi_devs)
        self._logger.debug("Total number of devices (post split): %s", len(devices))
        return devices

    async def generate_events_from_update(self, dev: AferoDevice) -> None:
        """Generate updates for a single device update."""
        devices = await self.split_devices([dev])
        self._logger.debug(
            "Received update for device %s. Generating %d events",
            dev.device_class,
            len(devices),
        )
        for device in devices:
            self._event_queue.put_nowait(
                AferoEvent(
                    type=EventType.RESOURCE_UPDATE_RESPONSE,
                    device_id=device.id,
                    device=device,
                    force_forward=False,
                )
            )

    async def generate_events_from_data(self, data: list[dict[Any, str]]) -> None:
        """Process the raw Afero IoT data for emitting.

        :param data: Raw data from Afero IoT
        """
        processed_ids = []
        skipped_ids = []
        devices = await self.generate_devices_from_data(data)
        self._event_queue.put_nowait(
            AferoEvent(
                type=EventType.POLLED_DATA,
                polled_data=data,
                force_forward=False,
            )
        )
        self._event_queue.put_nowait(
            AferoEvent(
                type=EventType.POLLED_DEVICES,
                polled_devices=devices,
                force_forward=False,
            )
        )
        for device in devices:
            event_type = EventType.RESOURCE_UPDATED
            if device.id not in self._bridge.tracked_devices:
                event_type = EventType.RESOURCE_ADDED
            self._event_queue.put_nowait(
                AferoEvent(
                    type=event_type,
                    device_id=device.id,
                    device=device,
                    force_forward=False,
                )
            )
            processed_ids.append(device.id)
        # Handle devices that did not report in from the API
        for dev_id in self._bridge.tracked_devices:
            if dev_id not in processed_ids + skipped_ids:
                self._event_queue.put_nowait(
                    AferoEvent(type=EventType.RESOURCE_DELETED, device_id=dev_id)
                )
                self._bridge.remove_device(dev_id)

    async def perform_poll(self) -> None:
        """Poll Afero IoT and generate the required events."""
        try:
            data = await self.gather_data()
        except Exception:  # noqa: BLE001
            self._status = EventStreamStatus.DISCONNECTED
            self.emit(EventType.DISCONNECTED)
        else:
            try:
                await self.generate_events_from_data(data)
            except Exception:
                self._logger.exception("Unable to process Afero IoT data. %s", data)
            self._first_poll_completed = True

    async def __event_reader(self) -> None:
        """Poll the current states."""
        self._status = EventStreamStatus.CONNECTING
        while True:
            await self.perform_poll()
            await asyncio.sleep(self._polling_interval)

    async def process_event(self):
        """Process a single event in the queue."""
        try:
            event: AferoEvent = await self._event_queue.get()
            self.emit(event["type"], event)
        except Exception:
            self._logger.exception("Unhandled exception. Please open a bug report")

    async def __event_processor(self) -> None:
        """Process the Afero IoT devices."""
        while True:
            await self.process_event()



"""Handles generic information related to all devices."""

from dataclasses import dataclass, field

from . import features
from .resource import DeviceInformation
from .sensor import AferoBinarySensor, AferoSensor


@dataclass(kw_only=True)
class StandardMixin:
    """Mixin for standard properties and methods."""

    _id: str  # ID used when interacting with Afero
    available: bool
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    split_identifier: str | None = None

    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=dict
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=dict
    )
    sensors: dict[str, AferoSensor] = field(default_factory=dict)
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=dict)
    # Defined at initialization
    instances: dict = field(default_factory=dict, repr=False, init=False)

    def __post_init__(self):
        """Configure the available instances."""
        instances = {}
        for function in self.device_information.functions or []:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem."""
        return self.instances.get(elem, None)

    @property
    def id(self):
        """ID for the device (split or normal)."""
        return self._id

    @property
    def instance(self):
        """Instance for the split device."""
        if self.split_identifier:
            return self.id.rsplit(f"-{self.split_identifier}-", 1)[1]
        return None

    @property
    def update_id(self) -> str:
        """ID used when sending updates to Afero API."""
        if self.split_identifier:
            return self.id.rsplit(f"-{self.split_identifier}-", 1)[0]
        return self.id



"""Representation of an Afero Security System Sensor (derived) and its corresponding updates."""

from dataclasses import dataclass, field

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class SecuritySystemSensor(StandardMixin):
    """Representation of a security system sensor."""

    type: ResourceTypes = ResourceTypes.SECURITY_SYSTEM_SENSOR

    config_key: str | None = None

    # Override the instance property to return an int
    @property
    def instance(self):
        """Instance for the split device."""
        return int(self._id.rsplit(f"-{self.split_identifier}-", 1)[1])


@dataclass
class SecuritySystemSensorPut:
    """States that can be updated for a Security System Sensor."""

    sensor_config: features.SecuritySensorConfigFeature | None = field(
        default_factory=dict, repr=False, init=False
    )



"""Test ThermostatController"""

import logging

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers.device import AferoBinarySensor
from aioafero.v1.controllers.thermostat import (
    ThermostatController,
    features,
    get_supported_modes,
)

from .. import utils

thermostat = utils.create_devices_from_data("thermostat.json")[0]
thermostat_id = "cc770a99-25da-4888-8a09-2a569da5be08"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.thermostats


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller[thermostat.id]
    assert dev.id == thermostat_id
    assert dev.current_temperature.temperature == 18.3
    assert dev.fan_mode == features.ModeFeature(
        mode="auto", modes={"on", "auto", "intermittent"}
    )
    assert dev.fan_running is False
    assert dev.hvac_action == "off"
    assert dev.hvac_mode == features.HVACModeFeature(
        mode="heat",
        previous_mode="heat",
        modes={"off", "heat", "auto", "fan", "cool"},
        supported_modes={"off", "heat", "fan"},
    )
    assert dev.safety_max_temp == features.TargetTemperatureFeature(
        value=36, min=29.5, max=37, step=0.5, instance="safety-mode-max-temp"
    )
    assert dev.safety_min_temp == features.TargetTemperatureFeature(
        value=4, min=4, max=13, step=0.5, instance="safety-mode-min-temp"
    )
    assert dev.target_temperature_auto_cooling == features.TargetTemperatureFeature(
        value=26.5, step=0.5, min=10, max=37, instance="auto-cooling-target"
    )
    assert dev.target_temperature_auto_heating == features.TargetTemperatureFeature(
        value=18.5, step=0.5, min=4, max=32, instance="auto-heating-target"
    )
    assert dev.target_temperature_cooling == features.TargetTemperatureFeature(
        value=26.5, step=0.5, min=10, max=37, instance="cooling-target"
    )
    assert dev.target_temperature_heating == features.TargetTemperatureFeature(
        value=18, step=0.5, min=4, max=32, instance="heating-target"
    )
    assert dev.sensors == {}
    assert dev.binary_sensors == {
        "filter-replacement|None": AferoBinarySensor(
            id="filter-replacement|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            current_value="not-needed",
            _error="replacement-needed",
            instance=None,
        ),
        "max-temp-exceeded|None": AferoBinarySensor(
            id="max-temp-exceeded|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            current_value="normal",
            _error="alerting",
            instance=None,
        ),
        "min-temp-exceeded|None": AferoBinarySensor(
            id="min-temp-exceeded|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            current_value="normal",
            _error="alerting",
            instance=None,
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            functionClass="current-fan-state", value="on", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="fan-mode", value="on", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="temperature", value=22, lastUpdateTime=0, functionInstance="auto-heating-target"
        ),
        AferoState(
            functionClass="temperature", value=22.5, lastUpdateTime=0, functionInstance="auto-cooling-target"
        ),
        AferoState(
            functionClass="temperature", value=17, lastUpdateTime=0, functionInstance="heating-target"
        ),
        AferoState(
            functionClass="temperature", value=18, lastUpdateTime=0, functionInstance="cooling-target"
        ),
        AferoState(
            functionClass="temperature", value=19, lastUpdateTime=0, functionInstance="current-temp"
        ),
        AferoState(
            functionClass="temperature", value=35, lastUpdateTime=0, functionInstance="safety-mode-max-temp"
        ),
        AferoState(
            functionClass="temperature", value=32, lastUpdateTime=0, functionInstance="safety-mode-min-temp"
        ),
        AferoState(
            functionClass="mode", value="cool", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="current-system-state", value="cooling", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller[thermostat_id]
    assert dev.current_temperature.temperature == 19
    assert dev.fan_running is True
    assert dev.fan_mode.mode == "on"
    assert dev.hvac_action == "cooling"
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "heat"
    assert dev.safety_max_temp.value == 35
    assert dev.safety_min_temp.value == 32
    assert dev.target_temperature_auto_heating.value == 22
    assert dev.target_temperature_auto_cooling.value == 22.5
    assert dev.target_temperature_heating.value == 17
    assert dev.target_temperature_cooling.value == 18
    assert dev.available is False
    assert updates == {
        "temperature-safety-mode-max-temp",
        "temperature-heating-target",
        "temperature-auto-cooling-target",
        "temperature-cooling-target",
        "current-fan-state",
        "temperature-current-temp",
        "current-system-state",
        "available",
        "temperature-auto-heating-target",
        "mode",
        "temperature-safety-mode-min-temp",
        "fan-mode",
    }


@pytest.mark.asyncio
async def test_update_elem_no_prev_mode_change(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = "off"
    mocked_controller[thermostat_id].hvac_mode.previous_mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            functionClass="mode", value="cool", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="temperature-units", value="fahrenheit", lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller[thermostat_id]
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "heat"
    assert updates == {"mode"}


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            functionClass="current-fan-state", value="off", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="fan-mode", value="auto", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="temperature", value=18.5, lastUpdateTime=0, functionInstance="auto-heating-target"
        ),
        AferoState(
            functionClass="temperature", value=26.5, lastUpdateTime=0, functionInstance="auto-cooling-target"
        ),
        AferoState(
            functionClass="temperature", value=18, lastUpdateTime=0, functionInstance="heating-target"
        ),
        AferoState(
            functionClass="temperature", value=26.5, lastUpdateTime=0, functionInstance="cooling-target"
        ),
        AferoState(
            functionClass="temperature", value=18.3, lastUpdateTime=0, functionInstance="current-temp"
        ),
        AferoState(
            functionClass="temperature", value=36, lastUpdateTime=0, functionInstance="safety-mode-max-temp"
        ),
        AferoState(
            functionClass="temperature", value=4, lastUpdateTime=0, functionInstance="safety-mode-min-temp"
        ),
        AferoState(
            functionClass="mode", value="heat", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="current-system-state", value="off", lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="available", value=True, lastUpdateTime=0, functionInstance=None
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller[thermostat_id].hvac_mode.supported_modes.add("cool")
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        thermostat_id,
        hvac_mode="cool",
        safety_max_temp=35,
        safety_min_temp=8,
        target_temperature_auto_heating=22,
        target_temperature_auto_cooling=22.5,
        target_temperature_heating=17,
        target_temperature_cooling=18,
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.fan_mode.mode == "auto"
    assert dev.fan_running is False
    assert dev.hvac_mode.mode == "cool"
    assert dev.safety_max_temp.value == 35
    assert dev.safety_min_temp.value == 8
    assert dev.target_temperature_auto_heating.value == 22
    assert dev.target_temperature_auto_cooling.value == 22.5
    assert dev.target_temperature_heating.value == 17
    assert dev.target_temperature_cooling.value == 18


@pytest.mark.asyncio
async def test_set_state_in_f_force_c(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller[thermostat_id].hvac_mode.supported_modes.add("cool")
    await mocked_controller.set_state(
        thermostat_id,
        hvac_mode="cool",
        safety_max_temp=35,
        safety_min_temp=8,
        target_temperature_auto_heating=22,
        target_temperature_auto_cooling=22.5,
        target_temperature_heating=17,
        target_temperature_cooling=18,
        is_celsius=True,
    )
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.fan_mode.mode == "auto"
    assert dev.fan_running is False
    assert dev.hvac_mode.mode == "cool"
    assert dev.safety_max_temp.value == 35
    assert dev.safety_min_temp.value == 8
    assert dev.target_temperature_auto_heating.value == 22
    assert dev.target_temperature_auto_cooling.value == 22.5
    assert dev.target_temperature_heating.value == 17
    assert dev.target_temperature_cooling.value == 18


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "current_mode, prev_mode, params, expected_result, expected_messages",
    [
        # Testing target_temp / cooling
        (
            "cool",
            "cool",
            {"target_temperature": 25},
            {"instance": "cooling-target", "value": 25},
            [],
        ),
        # Testing target_temp / cooling
        (
            "heat",
            "heat",
            {"target_temperature": 24},
            {"instance": "heating-target", "value": 24},
            [],
        ),
        # Testing changing mode
        (
            "heat",
            "heat",
            {"target_temperature": 24, "hvac_mode": "cool"},
            {"instance": "cooling-target", "value": 24, "mode": "cool"},
            [],
        ),
        # Testing target_temp / debug message
        (
            "off",
            "off",
            {"target_temperature": 25},
            [],
            ["Unable to set the target temperature due to the active mode: off"],
        ),
        # Invalid fan mode
        (
            "auto",
            "cool",
            {"fan_mode": "beans"},
            [],
            ["Unknown fan mode beans. Available modes: auto, intermittent, on"],
        ),
        # Invalid hvac mode
        (
            "auto",
            "cool",
            {"hvac_mode": "beans"},
            [],
            ["Unknown hvac mode beans. Available modes: cool, fan, heat, off"],
        ),
    ],
)
async def test_set_state_hvac_generics(
    current_mode,
    prev_mode,
    params,
    expected_result,
    expected_messages,
    mocked_controller,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = current_mode
    mocked_controller[thermostat_id].hvac_mode.supported_modes.add("cool")
    mocked_controller[thermostat_id].hvac_mode.previous_mode = prev_mode
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(thermostat_id, **params)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    if expected_result:
        if "mode" in expected_result:
            assert dev.hvac_mode.mode == expected_result["mode"]
        if expected_result["instance"] == "heating-target":
            assert dev.target_temperature_heating.value == expected_result["value"]
        elif expected_result["instance"] == "cooling-target":
            assert dev.target_temperature_cooling.value == expected_result["value"]
    for message in expected_messages:
        assert message in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "fan_mode, hvac_mode, params, expected_states, expected_messages",
    [
        (
            "off",
            "off",
            {"fan_mode": "auto"},
            {"fan-mode": "auto", "mode": "fan"},
            [],
        ),
        (
            "on",
            "fan",
            {"fan_mode": "auto"},
            {"fan-mode": "auto", "mode": "fan"},
            [],
        ),
        (
            "off",
            "off",
            {"fan_mode": "bad"},
            [],
            ["Unknown fan mode bad. Available modes: auto, intermittent, on"],
        ),
    ],
)
async def test_set_state_fan_generics(
    fan_mode,
    hvac_mode,
    params,
    expected_states,
    expected_messages,
    mocked_controller,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    mocked_controller[thermostat_id].hvac_mode.mode = hvac_mode
    mocked_controller[thermostat_id].fan_mode.mode = fan_mode
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(thermostat_id, **params)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    if expected_states:
        assert dev.fan_mode.mode == expected_states["fan-mode"]
        assert dev.hvac_mode.mode == expected_states["mode"]
    for message in expected_messages:
        assert message in caplog.text


@pytest.mark.asyncio
async def test_set_fan_mode(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[thermostat_id].fan_mode.mode = "off"
    await mocked_controller.set_fan_mode(thermostat_id, "on")
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.fan_mode.mode == "on"
    assert dev.hvac_mode.mode == "fan"


@pytest.mark.asyncio
async def test_set_hvac_mode(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller[thermostat_id].hvac_mode.supported_modes.add("cool")
    await mocked_controller.set_hvac_mode(thermostat_id, "cool")
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.hvac_mode.mode == "cool"


@pytest.mark.asyncio
async def test_set_target_temperature(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller[thermostat_id].target_temperature_heating.value = 20
    await mocked_controller.set_target_temperature(thermostat_id, 21)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.target_temperature_heating.value == 21


@pytest.mark.asyncio
async def test_set_temperature_range(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(thermostat)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    mocked_controller[thermostat_id].hvac_mode.mode = "auto"
    mocked_controller[thermostat_id].target_temperature_auto_heating.value = 20
    mocked_controller[thermostat_id].target_temperature_auto_cooling.value = 21
    await mocked_controller.set_temperature_range(thermostat_id, 21, 22)
    await mocked_controller._bridge.async_block_until_done()
    dev = mocked_controller[thermostat_id]
    assert dev.target_temperature_auto_heating.value == 21
    assert dev.target_temperature_auto_cooling.value == 22


standard = {"heat", "cool", "off", "fan", "auto"}


@pytest.mark.parametrize(
    "system_type, modes, expected",
    [
        ("cool-beans", {"heat", "cool", "beans"}, {"beans"}),
        ("1-compressor-heat-pump-1-aux-on-cool-boiler-aux", standard, standard),
        ("1-stage-cooling-conventional", standard, {"cool", "off", "fan"}),
        ("1-stage-heating-conventional-boiler", standard, {"heat", "off", "fan"}),
    ],
)
def test_get_supported_modes(system_type, modes, expected):
    assert get_supported_modes(system_type, modes) == expected


async def set_state_invalid_device(mocked_controller):
    assert await mocked_controller.set_state("invalid", hvac_mode="cool") is None



"""Test LockController"""

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.lock import LockController, features

from .. import utils

lock = utils.create_devices_from_data("door-lock-TBD.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.locks


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(lock)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "698e8a63-e8cb-4335-ba6b-83ca69d378f2"
    assert dev.position == features.CurrentPositionFeature(
        position=features.CurrentPositionEnum.LOCKED
    )


@pytest.mark.asyncio
async def test_lock(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(lock)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.lock(lock.id)
    await mocked_controller._bridge.async_block_until_done()
    assert mocked_controller.items[0].position.position == features.CurrentPositionEnum.LOCKING


@pytest.mark.asyncio
async def test_unlock(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(lock)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    await mocked_controller.unlock(lock.id)
    await mocked_controller._bridge.async_block_until_done()
    assert mocked_controller.items[0].position.position == features.CurrentPositionEnum.UNLOCKING


@pytest.mark.asyncio
async def test_empty_update(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(lock)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    update = utils.create_devices_from_data("door-lock-TBD.json")[0]
    updates = await mocked_controller.update_elem(update)
    assert updates == set()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "value, expected, expected_updates",
    [
        ("locking", features.CurrentPositionEnum.LOCKING, {"position"}),
        ("unlocking", features.CurrentPositionEnum.UNLOCKING, {"position"}),
        ("not-a-state", features.CurrentPositionEnum.UNKNOWN, {"position"}),
    ],
)
async def test_update_elem(value, expected, expected_updates, mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(lock)]
    )
    await mocked_controller._bridge.async_block_until_done()
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available
    dev_update = utils.create_devices_from_data("door-lock-TBD.json")[0]
    new_states = [
        AferoState(
            functionClass="lock-control", value=value, lastUpdateTime=0, functionInstance=None
        ),
        AferoState(
            functionClass="available", value=False, lastUpdateTime=0, functionInstance=None
        ),
    ]
    expected_updates.add("available")
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    assert dev.position.position == expected
    assert not dev.available
    assert updates == expected_updates


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller._bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(lock)]
    )
    await mocked_controller._bridge.async_block_until_done()
    await mocked_controller.set_state(lock.id)


@pytest.mark.asyncio
async def test_lock_emitting(bridge):
    dev_update = utils.create_devices_from_data("door-lock-TBD.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    await bridge.async_block_until_done()
    assert len(bridge.locks._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    await bridge.async_block_until_done()
    assert len(bridge.locks._items) == 1
    assert not bridge.locks._items[dev_update.id].available



"""Models for all currently-tracked Afero IoT device classes."""

__all__ = [
    "AferoBinarySensor",
    "AferoSensor",
    "Device",
    "DeviceInformation",
    "ExhaustFan",
    "ExhaustFanPut",
    "Fan",
    "FanPut",
    "Light",
    "LightPut",
    "Lock",
    "LockPut",
    "PortableAC",
    "PortableACPut",
    "ResourceTypes",
    "SecuritySystem",
    "SecuritySystemKeypad",
    "SecuritySystemKeypadPut",
    "SecuritySystemPut",
    "SecuritySystemSensor",
    "SecuritySystemSensorPut",
    "Switch",
    "SwitchPut",
    "Thermostat",
    "ThermostatPut",
    "Valve",
    "ValvePut",
]


from .device import Device
from .exhaust_fan import ExhaustFan, ExhaustFanPut
from .fan import Fan, FanPut
from .light import Light, LightPut
from .lock import Lock, LockPut
from .portable_ac import PortableAC, PortableACPut
from .resource import DeviceInformation, ResourceTypes
from .security_system import SecuritySystem, SecuritySystemPut
from .security_system_keypad import SecuritySystemKeypad, SecuritySystemKeypadPut
from .security_system_sensor import SecuritySystemSensor, SecuritySystemSensorPut
from .sensor import AferoBinarySensor, AferoSensor
from .switch import Switch, SwitchPut
from .thermostat import Thermostat, ThermostatPut
from .valve import Valve, ValvePut



"""Constants for accessing Afero API."""

from string import Template
from typing import Final

AFERO_CLIENTS: Final[dict[str, dict[str, str]]] = {
    "hubspace": {
        "API_DATA_HOST": "semantics2.afero.net",
        "API_HOST": "api2.afero.net",
        "AUTH_DEFAULT_CLIENT_ID": "hubspace_android",
        "AUTH_DEFAULT_REDIRECT_URI": "hubspace-app://loginredirect",
        "AUTH_OPENID_HOST": "accounts.hubspaceconnect.com",
        "AUTH_REALM": "thd",
    },
    "myko": {
        "API_DATA_HOST": "semantics2.sxz2xlhh.afero.net",
        "API_HOST": "api2.sxz2xlhh.afero.net",
        "AUTH_DEFAULT_CLIENT_ID": "kfi_android",
        "AUTH_DEFAULT_REDIRECT_URI": "kfi-app://loginredirect",
        "AUTH_OPENID_HOST": "accounts.mykoapp.com",
        "AUTH_REALM": "kfi",
    },
}


AFERO_GENERICS: Final[dict[str, str]] = {
    # Generics
    "DEFAULT_USERAGENT": Template(
        "Mozilla/5.0 (Linux; Android 15; ${client_name} Build/test; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/138.0.7204.63 Mobile Safari/537.36"
    ),
    # API Endpoints
    "API_DEVICE_ENDPOINT": "/v1/accounts/{}/metadevices",
    "API_DEVICE_STATE_ENDPOINT": "/v1/accounts/{}/metadevices/{}/state",
    "API_DEVICE_VERSIONS_ENDPOINT": "/v1/accounts/{}/devices/{}/versions",
    # Auth endpoints
    "AUTH_OPENID_ENDPOINT": "/protocol/openid-connect/auth",
    "AUTH_CODE_ENDPOINT": "/login-actions/authenticate",
    "AUTH_TOKEN_ENDPOINT": "/protocol/openid-connect/token",
    "ACCOUNT_ID_ENDPOINT": "/v1/users/me",
}

MAX_RETRIES: Final[int] = 3

# Version polling interval in seconds (6 hours)
VERSION_POLL_INTERVAL_SECONDS = 6 * 3600



"""Representation of an Afero Exhaust Fan and its corresponding updates."""

from dataclasses import dataclass, field

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class ExhaustFan(StandardMixin):
    """Representation of an Afero Exhaust Fan."""

    type: ResourceTypes = ResourceTypes.EXHAUST_FAN


@dataclass
class ExhaustFanPut:
    """States that can be updated for an exhaust fan."""

    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )



"""Representation of an Afero Security System and its corresponding updates."""

from dataclasses import dataclass, field

from aioafero.v1.models import features

from .resource import ResourceTypes
from .standard_mixin import StandardMixin


@dataclass(kw_only=True)
class SecuritySystem(StandardMixin):
    """Representation of an Afero Security Panel."""

    type: ResourceTypes = ResourceTypes.SECURITY_SYSTEM

    alarm_state: features.ModeFeature | None = None
    siren_action: features.SecuritySensorSirenFeature | None = None

    @property
    def supports_away(self) -> bool:
        """States if the panel supports away mode."""
        return "arm-away" in self.alarm_state.modes

    @property
    def supports_arm_bypass(self) -> bool:
        """States if the panel supports arm-bypass mode."""
        return False

    @property
    def supports_home(self) -> bool:
        """States if the panel supports home mode."""
        return "arm-stay" in self.alarm_state.modes

    @property
    def supports_night(self) -> bool:
        """States if the panel supports night mode."""
        return False

    @property
    def supports_vacation(self) -> bool:
        """States if the panel supports vacation mode."""
        return False

    @property
    def supports_trigger(self) -> bool:
        """States if the panel supports manually triggering."""
        return "alarming-sos" in self.alarm_state.modes


@dataclass
class SecuritySystemPut:
    """States that can be updated for a Security System."""

    alarm_state: features.ModeFeature | None = None
    siren_action: features.SecuritySensorSirenFeature | None = None
    disarm_pin: features.SecuritySystemDisarmPin | None = None
    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=dict, repr=False, init=False
    )



import pytest

from aioafero import AferoState, get_afero_device
from aioafero.v1.controllers import event
from aioafero.v1.controllers.device import DeviceController
from aioafero.v1.models.resource import DeviceInformation
from aioafero.v1.models.sensor import AferoBinarySensor, AferoSensor

from .. import utils
from ..utils import modify_state
from .test_event import security_system_callback

a21_light = utils.create_devices_from_data("light-a21.json")[0]
zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]
freezer = utils.create_devices_from_data("freezer.json")[0]
door_lock = utils.create_devices_from_data("door-lock-TBD.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    return mocked_bridge.devices


@pytest.fixture
def security_system_devices():
    data = utils.get_raw_dump("security-system-raw.json")
    devices = [get_afero_device(dev) for dev in data]
    assert len(devices) == 2
    multi_devs = []
    for dev in devices:
        multi_devs.extend(security_system_callback(dev).split_devices)
    devices.extend(multi_devs)
    assert len(devices) == 5
    return devices


@pytest.mark.asyncio
async def test_initialize_not_needed(mocked_controller):
    await mocked_controller.initialize()
    await mocked_controller.initialize()


@pytest.mark.asyncio
async def test_initialize_a21(mocked_controller):
    await mocked_controller.initialize_elem(a21_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == a21_light.id
    assert dev.available is True
    assert dev.device_information == DeviceInformation(
        device_class=a21_light.device_class,
        default_image=a21_light.default_image,
        default_name=a21_light.default_name,
        manufacturer=a21_light.manufacturerName,
        model=a21_light.model,
        name=a21_light.friendly_name,
        parent_id=a21_light.device_id,
        wifi_mac="b31d2f3f-86f6-4e7e-b91b-4fbc161d410d",
        ble_mac="9c70c759-1d54-4f61-a067-bb4294bef7ae",
        functions = a21_light.functions,
        children=a21_light.children,
    )
    assert dev.sensors == {
        "wifi-rssi": AferoSensor(
            id="wifi-rssi",
            owner="dd883754-e9f2-4c48-b755-09bf6ce776be",
            value=-50,
            instance=None,
            unit="dB",
        )
    }
    assert dev.binary_sensors == {}


@pytest.mark.asyncio
async def test_initialize_door_lock(mocked_controller):
    await mocked_controller.initialize_elem(door_lock)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == door_lock.id
    assert dev.available is True
    assert dev.device_information == DeviceInformation(
        device_class=door_lock.device_class,
        default_image=door_lock.default_image,
        default_name=door_lock.default_name,
        manufacturer=door_lock.manufacturerName,
        model=door_lock.model,
        name=door_lock.friendly_name,
        parent_id=door_lock.device_id,
        wifi_mac="6f6882f2-b35f-451f-bab1-4feafe33dbb3",
        ble_mac="1392f7cb-e23a-470e-b803-6be2e48ce5c0",
        functions = door_lock.functions,
        children=door_lock.children,
    )
    assert dev.sensors == {
        "battery-level": AferoSensor(
            id="battery-level",
            owner="698e8a63-e8cb-4335-ba6b-83ca69d378f2",
            value=80,
            unit="%",
            instance=None,
        ),
    }
    assert dev.binary_sensors == {}


@pytest.mark.asyncio
async def test_initialize_binary_sensors(mocked_controller):
    await mocked_controller.initialize_elem(freezer)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == freezer.id
    assert dev.available is True
    assert dev.device_information == DeviceInformation(
        device_class=freezer.device_class,
        default_image=freezer.default_image,
        default_name=freezer.default_name,
        manufacturer=freezer.manufacturerName,
        model=freezer.model,
        name=freezer.friendly_name,
        parent_id=freezer.device_id,
        wifi_mac="351cccd0-87ff-41b3-b18c-568cf781d56d",
        ble_mac="c2e189e8-c80c-4948-9492-14ac390f480d",
        functions = freezer.functions,
        children=freezer.children,
    )
    assert dev.sensors == {
        "wifi-rssi": AferoSensor(
            id="wifi-rssi",
            owner="eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7",
            value=-71,
            instance=None,
            unit="dB",
        )
    }
    assert dev.binary_sensors == {
        "error|freezer-high-temperature-alert": AferoBinarySensor(
            id="error|freezer-high-temperature-alert",
            owner="eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7",
            current_value="normal",
            _error="alerting",
            instance="freezer-high-temperature-alert",
        ),
        "error|fridge-high-temperature-alert": AferoBinarySensor(
            id="error|fridge-high-temperature-alert",
            owner="eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7",
            current_value="alerting",
            _error="alerting",
            instance="fridge-high-temperature-alert",
        ),
        "error|mcu-communication-failure": AferoBinarySensor(
            id="error|mcu-communication-failure",
            owner="eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7",
            current_value="normal",
            _error="alerting",
            instance="mcu-communication-failure",
        ),
        "error|temperature-sensor-failure": AferoBinarySensor(
            id="error|temperature-sensor-failure",
            owner="eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7",
            current_value="normal",
            _error="alerting",
            instance="temperature-sensor-failure",
        ),
    }


@pytest.mark.parametrize(
    "filename, expected",
    [
        (
            "hs_data.json",
            [
                "82a71f6f-3689-424c-9f12-db0552b2726f",
                "14ff1a25-c945-4827-bfa2-e292297d27c3",
                "3624a09a-ba5e-4f27-846a-916a0527dd5b",
                "5e87cd46-4b4e-4ac0-9a13-f65ab9546001",
            ],
        ),
        (
            "water-timer.json",
            [
                "86114564-7acd-4542-9be9-8fd798a22b06",
            ],
        ),
    ],
)
def test_get_filtered_devices(filename, expected, mocked_controller):
    data = utils.create_devices_from_data(filename)
    res = mocked_controller.get_filtered_devices(data)
    actual_devs = [x.device_id for x in res]
    assert len(actual_devs) == len(expected)
    for key in expected:
        assert key in actual_devs


def test_get_filtered_devices_security_system(
    mocked_controller, security_system_devices
):
    assert len(mocked_controller.get_filtered_devices(security_system_devices)) == 5
    assert security_system_devices[0].friendly_name == "Main Keypad"
    assert security_system_devices[1].friendly_name == "Helms Deep"
    assert security_system_devices[2].friendly_name == "Helms Deep - Sensor 1"
    assert security_system_devices[3].friendly_name == "Helms Deep - Sensor 2"
    assert security_system_devices[4].friendly_name == "Helms Deep - Sensor 4"


@pytest.mark.asyncio
async def test__process_polled_devices(
    mocked_controller, security_system_devices, mocker
):
    evt = event.AferoEvent(
        type=event.EventType.POLLED_DEVICES, polled_devices=security_system_devices
    )
    await mocked_controller._process_polled_devices(evt["type"], evt)
    assert len(mocked_controller.items) == 5


@pytest.mark.asyncio
async def test_update_elem_sensor(mocked_controller):
    await mocked_controller.initialize_elem(a21_light)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == a21_light.id
    dev_update: utils.AferoDevice = utils.create_devices_from_data("light-a21.json")[0]
    unavail = AferoState(
        functionClass="available",
        value=False,
    )
    utils.modify_state(dev_update, unavail)
    rssi = AferoState(
        functionClass="wifi-rssi",
        value="40db",
    )
    utils.modify_state(dev_update, rssi)
    updates = await mocked_controller.update_elem(dev_update)
    assert dev.available is False
    assert dev.sensors["wifi-rssi"].value == 40
    assert updates == {"available", "sensor-wifi-rssi"}


@pytest.mark.asyncio
async def test_update_elem_binary_sensor(mocked_controller):
    await mocked_controller.initialize_elem(freezer)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == freezer.id
    dev_update: utils.AferoDevice = utils.create_devices_from_data("freezer.json")[0]
    temp_sensor_failure = AferoState(
        functionClass="error",
        functionInstance="temperature-sensor-failure",
        value="alerting",
    )
    utils.modify_state(dev_update, temp_sensor_failure)
    updates = await mocked_controller.update_elem(dev_update)
    assert dev.binary_sensors["error|temperature-sensor-failure"].value is True
    assert updates == {"binary-error|temperature-sensor-failure"}


@pytest.mark.asyncio
async def test_update_lifecycle(bridge, caplog):
    caplog.set_level(0)
    polled_devices = utils.create_devices_from_data("freezer.json")
    # Simulate a new poll with a new device
    polled_devices_event = event.AferoEvent(
        type=event.EventType.POLLED_DEVICES,
        polled_devices=polled_devices,
    )
    bridge.events.emit(polled_devices_event["type"], polled_devices_event)
    await bridge.async_block_until_done()
    assert len(bridge.devices.items) == 1
    assert bridge.devices._known_parents == {
        "596c120d-4e0d-4e33-ae9a-6330dcf2cbb5": "eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7"
    }
    assert (
        "Initializing friendly-device-0 [eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7] as a Device"
        in caplog.text
    )
    dev = bridge.devices["eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7"]
    assert dev.available
    assert dev.sensors["wifi-rssi"].value == -71
    assert dev.binary_sensors["error|temperature-sensor-failure"].value is False
    # Simulate an update
    modify_state(
        polled_devices[0],
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    modify_state(
        polled_devices[0],
        AferoState(
            functionClass="wifi-rssi",
            functionInstance=None,
            value=-42,
        ),
    )
    modify_state(
        polled_devices[0],
        AferoState(
            functionClass="error",
            functionInstance="temperature-sensor-failure",
            value="alerting",
        ),
    )
    polled_devices_event = event.AferoEvent(
        type=event.EventType.POLLED_DEVICES,
        polled_devices=polled_devices,
    )
    bridge.events.emit(polled_devices_event["type"], polled_devices_event)
    await bridge.async_block_until_done()
    assert len(bridge.devices.items) == 1
    assert bridge.devices._known_parents == {
        "596c120d-4e0d-4e33-ae9a-6330dcf2cbb5": "eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7"
    }
    dev = bridge.devices["eacfca4b-4f4b-4ee2-aa64-e1052fa9cea7"]
    assert not dev.available
    assert dev.sensors["wifi-rssi"].value == -42
    assert dev.binary_sensors["error|temperature-sensor-failure"].current_value
    # Simulate removal
    polled_devices_event = event.AferoEvent(
        type=event.EventType.POLLED_DEVICES,
        polled_devices=[],
    )
    bridge.events.emit(polled_devices_event["type"], polled_devices_event)
    await bridge.async_block_until_done()
    assert (
        "Device 596c120d-4e0d-4e33-ae9a-6330dcf2cbb5 was not polled. Removing"
        in caplog.text
    )
    assert len(bridge.devices.items) == 0
    assert bridge.devices._known_parents == {}


@pytest.mark.asyncio
async def test__process_update_response(mocked_controller, mocker):
    bridge = mocked_controller._bridge
    await bridge.events.generate_events_from_data(
        [utils.create_hs_raw_from_device(a21_light)]
    )
    await bridge.async_block_until_done()
    handle_event = mocker.patch.object(mocked_controller, "_handle_event", return_value=None)
    dev = utils.create_devices_from_data("light-a21.json")[0]
    await mocked_controller._process_update_response(
        event.EventType.RESOURCE_UPDATE_RESPONSE,
        event.AferoEvent(
            type=event.EventType.RESOURCE_UPDATE_RESPONSE,
            device=dev,
        ),
    )
    handle_event.assert_called_once()
    handle_event.reset_mock()
    dev.device_id = "beans"
    await mocked_controller._process_update_response(
        event.EventType.RESOURCE_UPDATE_RESPONSE,
        event.AferoEvent(
            type=event.EventType.RESOURCE_UPDATE_RESPONSE,
            device=dev,
        ),
    )
    handle_event.assert_not_called()


def test_instanced_known_devs(mocked_bridge):
    """Test that there is no overlap between sensor and binary_sensor IDs."""
    c1 = DeviceController(mocked_bridge)
    c2 = DeviceController(mocked_bridge)
    assert id(c1._known_parents) != id(c2._known_parents)
